/*
Copyright (c) 2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/


package traceccgui;

// jlayout.import.begin
// START OF JLAYOUT SECTION
// This section was generated automatically by the jlayout program.
// You should not modify this section.
// Any changes applied to this section will be lost if jlayout is run again.
import java.awt.BorderLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Dimension;
import java.awt.Container;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.ImageIcon;
import javax.swing.WindowConstants;
// END OF JLAYOUT SECTION
// jlayout.import.end

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.awt.event.WindowEvent;
import java.awt.Desktop;
import javax.swing.Box;
import javax.swing.JFileChooser;
import javax.swing.SwingUtilities;
import java.net.URI;
import java.net.URL;
import dirk_krause.jtools.StringTool;
import dirk_krause.jtools.GuiController;
import dirk_krause.jtools.GuiDialog;
import dirk_krause.jtools.ApplicationFrame;
import dirk_krause.jtools.DirectoryFilter;
import dirk_krause.jtools.SizeTracker;
import dirk_krause.jtools.OneDirectoryDropTransferHandler;
import dirk_krause.jtools.ComponentPositioner;
import dirk_krause.jtools.LogContentsForCommand;
import dirk_krause.jtools.LogWindow;
import dirk_krause.jtools.HelpWindow;


/**
 *
 * @author krause
 */
public class TraceccMainWindow extends ApplicationFrame
	implements ActionListener
{
  class DropHandler extends OneDirectoryDropTransferHandler {
    public DropHandler() {
      super();
    }
    @Override
    protected boolean processFilename(String s) {
      boolean back = false;
      File f = new File(s);
      if(f.isDirectory()) {
        setNewDirectory(s); back = true;
      }
      return back;
    }
  }

  class BackgroundThread implements Runnable {
      @Override
      public void run() {
          backgroundThreadFunction();
      }
  }

  class ConversionDone implements Runnable {
      private TraceccController t = null;
      private LogContentsForCommand lcfc = null;
      public ConversionDone(TraceccController mytc, LogContentsForCommand l) {
        t = mytc; lcfc = l;
      }
      @Override
      public void run() {
        conversionDone(t, lcfc);
      }
  }

  class StatusChange implements Runnable {
    private String st1 = null;
    private String st2 = null;
    private Color  col = null;
    public StatusChange(String s1, String s2, Color co) {
      st1 = s1; st2 = s2; col = co;
    }
    @Override
    public void run() {
      changeStatus(st1, st2, col);
    }
  }

  private TraceccController tc = null;
  private GuiController gc = null;
  private int b = 5;
  private Color cGreen = null;
  private Color cRed = null;
  private SizeTracker stDirectory = null;
  private SizeTracker stStatus = null;
  private SizeTracker stCurrentStatus = null;
  private DropHandler dh = null;
  private TraceccOptionsDialog tod = null;
  private GuiDialog guid = null;
  private LogWindow lw = null;
  private HelpWindow hw = null;

  /** Flag to indicate whether or not we have a background thread running */
  private boolean bRunning = false;

  /** Reference to thread */
  Thread thrBackground = null;

  public synchronized boolean isRunning() {
    return bRunning;
  }

  public synchronized void setRunning(boolean b, Thread t) {
    bRunning = b; thrBackground = t;
  }

  public synchronized Thread getThread() {
    return thrBackground;
  }

  private synchronized TraceccOptionsDialog getTod() {
    return tod;
  }

  private synchronized void setTod(TraceccOptionsDialog t) {
    tod = t;
  }

  private TraceccOptionsDialog useTod() {
    TraceccOptionsDialog back = getTod();
    if(back == null) {
      back = new TraceccOptionsDialog(this, tc, gc);
      ComponentPositioner.choosePosition(this, back, true);
      setTod(back);
      back.pack();
    }
    return back;
  }

  private synchronized LogWindow getLogWindow() {
      return lw;
  }

  private synchronized void setLogWindow(LogWindow l) {
      lw = l;
  }

  private LogWindow useLogWindow() {
      LogWindow back = getLogWindow();
      if(back == null) {
          back = new LogWindow(this, gc);
          ComponentPositioner.choosePosition(this, back, false);
          setLogWindow(back);
          back.pack();
      }
      return back;
  }

  private synchronized HelpWindow getHelpWindow() {
      return hw;
  }

  private synchronized void setHelpWindow(HelpWindow h) {
      hw = h;
  }

  private HelpWindow useHelpWindow() {
      HelpWindow back = getHelpWindow();
      if(back == null) {
          back = new HelpWindow(this, gc);
          ComponentPositioner.choosePosition(this, back, true);
          setHelpWindow(back);
          back.pack();
      }
      return back;
  }

  private synchronized GuiDialog getGuid() {
      return guid;
  }

  private synchronized void setGuid(GuiDialog g) {
      guid = g;
  }

  private GuiDialog useGuid() {
      GuiDialog g = getGuid();
      if(g == null) {
          g = new GuiDialog(this, gc);
          setGuid(g);
          g.pack();
          ComponentPositioner.choosePosition(this, g, true);
      }
      return g;
  }

// jlayout.variables.begin
// START OF JLAYOUT SECTION
// This section was generated automatically by the jlayout program.
// You should not modify this section.
// Any changes applied to this section will be lost if jlayout is run again.
	JButton bDirectory = null;
	JButton bExit = null;
	JButton bOptions = null;
	JButton bRun = null;
	JLabel lCurrentDir = null;
	JLabel lCurrentStatus = null;
	JLabel lDir = null;
	JLabel lIcon = null;
	JLabel lStatus = null;
	JMenuItem mAbout = null;
	JMenuItem mDirectory = null;
	JMenuItem mExit = null;
	JMenu mFile = null;
	JMenuItem mGui = null;
	JMenu mHelp = null;
	JMenuItem mLocalHelp = null;
	JMenuItem mOptions = null;
	JMenuItem mRun = null;
	JMenu mSetup = null;
	JMenuItem mWebHelp = null;
	JMenuBar mb = null;
	JPanel pContents = null;
	JToolBar tbButtons = null;
// END OF JLAYOUT SECTION
// jlayout.variables.end

// jlayout.setup.begin
// START OF JLAYOUT SECTION
// This section was generated automatically by the jlayout program.
// You should not modify this section.
// Any changes applied to this section will be lost if jlayout is run again.
	/**
	 * Find an image icon.
	 *
	 * @param n
	 * The name of the image icon file.
	 *
	 * @return
	 * The image icon.
	 */
	protected ImageIcon jLayoutImageIconFromName(String n)
	{
	  ImageIcon back = null;
	  try {
	    back = new ImageIcon(getClass().getResource(n));
	  }
	  catch(NullPointerException e) {
	    back = new ImageIcon(n);
	  }
	  return back;
	}
	/**
	 * Create GridBag constraints for a component.
	 *
	 * @param x
	 * The start grid column of the component.
	 *
	 * @param y
	 * The start grid line of the component.
	 *
	 * @param w
	 * The width of the component in grid columns.
	 *
	 * @param h
	 * The height of the component in grid lines.
	 *
	 * @param f
	 * The fill information (horizontal, vertical, both).
	 *
	 * @param wx
	 * The weight factor to grow the component horizontally
	 * when the container grows.
	 *
	 * @param wy
	 * The weight factor to grow the component vertically
	 * when the container grows.
	 */
	private GridBagConstraints jLayoutGbc(
	  int x, int y, int w, int h, int f, int a, double wx, double wy
	) {
	  GridBagConstraints back = new GridBagConstraints();
	  back.gridx = x;
	  back.gridy = y;
	  back.gridwidth = w;
	  back.gridheight = h;
	  back.fill = f;
	  back.anchor = a;
	  back.weightx = wx;
	  back.weighty = wy;
	  return back;
	}
	/**
	 * Create a dummy panel to ensure minimum distances.
	 *
	 * @param b
	 * The panels width and hight.
	 *
	 * @param rms
	 * A flag to indicate whether or not the panel must
	 * be set exactly to the given size (The panel can not grow).
	 */
	protected JPanel jLayoutCreateDistancePanel(int b, boolean rms)
	{
	  JPanel back = new JPanel();
	  Dimension d = new Dimension(b, b);
	  back.setMinimumSize(d);
	  back.setPreferredSize(d);
	  if(rms) {
	    back.setMaximumSize(d);
	  }
	  return back;
	}
	private void jLayoutSetup()
	{
	  // bDirectory 79
	  bDirectory = new JButton(jLayoutImageIconFromName(TGS.lt[4]));
	  bDirectory.setActionCommand(TGS.ac[0]);
	  bDirectory.setToolTipText(TGS.lt[47]);
	  bDirectory.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[47]
	  );
	  // bExit 82
	  bExit = new JButton(jLayoutImageIconFromName(TGS.lt[7]));
	  bExit.setActionCommand(TGS.ac[3]);
	  bExit.setToolTipText(TGS.lt[50]);
	  bExit.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[50]
	  );
	  // bOptions 80
	  bOptions = new JButton(jLayoutImageIconFromName(TGS.lt[5]));
	  bOptions.setActionCommand(TGS.ac[1]);
	  bOptions.setToolTipText(TGS.lt[48]);
	  bOptions.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[48]
	  );
	  // bRun 81
	  bRun = new JButton(jLayoutImageIconFromName(TGS.lt[6]));
	  bRun.setActionCommand(TGS.ac[2]);
	  bRun.setToolTipText(TGS.lt[49]);
	  bRun.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[49]
	  );
	  // lCurrentDir 112
	  lCurrentDir = new JLabel(TGS.lt[10]);
	  // lCurrentStatus 114
	  lCurrentStatus = new JLabel(TGS.lt[11]);
	  // lDir 111
	  lDir = new JLabel(TGS.lt[8]);
	  // lIcon 110
	  lIcon = new JLabel(jLayoutImageIconFromName(TGS.lt[12]));
	  // lStatus 113
	  lStatus = new JLabel(TGS.lt[9]);
	  lStatus.setForeground(cGreen);
	  // mAbout 61
	  mAbout = new JMenuItem(TGS.ac[7]);
	  mAbout.setToolTipText(TGS.lt[54]);
	  mAbout.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[54]
	  );
	  // mDirectory 23
	  mDirectory = new JMenuItem(TGS.ac[0]);
	  mDirectory.setToolTipText(TGS.lt[47]);
	  mDirectory.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[47]
	  );
	  // mExit 26
	  mExit = new JMenuItem(TGS.ac[3]);
	  mExit.setToolTipText(TGS.lt[50]);
	  mExit.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[50]
	  );
	  // mGui 50
	  mGui = new JMenuItem(TGS.ac[4]);
	  mGui.setToolTipText(TGS.lt[51]);
	  mGui.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[51]
	  );
	  // mLocalHelp 59
	  mLocalHelp = new JMenuItem(TGS.ac[5]);
	  mLocalHelp.setToolTipText(TGS.lt[52]);
	  mLocalHelp.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[52]
	  );
	  // mOptions 24
	  mOptions = new JMenuItem(TGS.ac[1]);
	  mOptions.setToolTipText(TGS.lt[48]);
	  mOptions.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[48]
	  );
	  // mRun 25
	  mRun = new JMenuItem(TGS.ac[2]);
	  mRun.setToolTipText(TGS.lt[49]);
	  mRun.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[49]
	  );
	  // mSetup 17
	  mSetup = new JMenu();
	  mSetup.setText(TGS.lt[2]);
	  {
	    mSetup.add(mGui);
	  }
	  // mWebHelp 60
	  mWebHelp = new JMenuItem(TGS.ac[6]);
	  mWebHelp.setToolTipText(TGS.lt[53]);
	  mWebHelp.getAccessibleContext().setAccessibleDescription(
	    TGS.lt[53]
	  );
	  // pContents 13
	  pContents = new JPanel();
	  {
	    GridBagLayout gbl = new GridBagLayout();
	    pContents.setLayout(gbl);
	    GridBagConstraints gbc = null;
	    JPanel dp = null;
	    dp = jLayoutCreateDistancePanel(b, false);
	    gbc = jLayoutGbc(
	      0, 0, 1, 1,
	      GridBagConstraints.BOTH, GridBagConstraints.CENTER,
	      0.5, 0.5
	    );
	    gbl.setConstraints(dp, gbc);
	    pContents.add(dp);
	    dp = jLayoutCreateDistancePanel(b, false);
	    gbc = jLayoutGbc(
	      0, 6, 1, 1,
	      GridBagConstraints.BOTH, GridBagConstraints.CENTER,
	      0.5, 0.5
	    );
	    gbl.setConstraints(dp, gbc);
	    pContents.add(dp);
	    dp = jLayoutCreateDistancePanel(b, false);
	    gbc = jLayoutGbc(
	      4, 0, 1, 1,
	      GridBagConstraints.BOTH, GridBagConstraints.CENTER,
	      0.5, 0.5
	    );
	    gbl.setConstraints(dp, gbc);
	    pContents.add(dp);
	    dp = jLayoutCreateDistancePanel(b, false);
	    gbc = jLayoutGbc(
	      4, 6, 1, 1,
	      GridBagConstraints.BOTH, GridBagConstraints.CENTER,
	      0.5, 0.5
	    );
	    gbl.setConstraints(dp, gbc);
	    pContents.add(dp);
	    dp = jLayoutCreateDistancePanel(b, true);
	    gbc = jLayoutGbc(
	      2, 0, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.CENTER,
	      0.0, 0.0
	    );
	    gbl.setConstraints(dp, gbc);
	    pContents.add(dp);
	    dp = jLayoutCreateDistancePanel(b, true);
	    gbc = jLayoutGbc(
	      0, 2, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.CENTER,
	      0.0, 0.0
	    );
	    gbl.setConstraints(dp, gbc);
	    pContents.add(dp);
	    dp = jLayoutCreateDistancePanel(b, true);
	    gbc = jLayoutGbc(
	      0, 4, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.CENTER,
	      0.0, 0.0
	    );
	    gbl.setConstraints(dp, gbc);
	    pContents.add(dp);
	    gbc = jLayoutGbc(
	      1, 1, 3, 1,
	      GridBagConstraints.NONE, GridBagConstraints.CENTER,
	      0.0, 0.0
	    );
	    gbl.setConstraints(lIcon, gbc);
	    pContents.add(lIcon);
	    gbc = jLayoutGbc(
	      1, 3, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.EAST,
	      0.0, 0.0
	    );
	    gbl.setConstraints(lDir, gbc);
	    pContents.add(lDir);
	    gbc = jLayoutGbc(
	      3, 3, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.WEST,
	      0.0, 0.0
	    );
	    gbl.setConstraints(lCurrentDir, gbc);
	    pContents.add(lCurrentDir);
	    gbc = jLayoutGbc(
	      1, 5, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.EAST,
	      0.0, 0.0
	    );
	    gbl.setConstraints(lStatus, gbc);
	    pContents.add(lStatus);
	    gbc = jLayoutGbc(
	      3, 5, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.WEST,
	      0.0, 0.0
	    );
	    gbl.setConstraints(lCurrentStatus, gbc);
	    pContents.add(lCurrentStatus);
	  }
	  // tbButtons 12
	  tbButtons = new JToolBar();
	  {
	    tbButtons.add(bDirectory);
	    tbButtons.add(bOptions);
	    tbButtons.add(bRun);
	    tbButtons.add(bExit);
	  }
	  // mFile 16
	  mFile = new JMenu();
	  mFile.setText(TGS.lt[1]);
	  {
	    mFile.add(mDirectory);
	    mFile.add(mOptions);
	    mFile.add(mRun);
	    mFile.add(mExit);
	  }
	  // mHelp 19
	  mHelp = new JMenu();
	  mHelp.setText(TGS.lt[3]);
	  {
	    mHelp.add(mLocalHelp);
	    mHelp.add(mWebHelp);
	    mHelp.add(mAbout);
	  }
	  // mb 10
	  mb = new JMenuBar();
	  {
	    mb.add(mFile);
	    mb.add(mSetup);
	    mb.add(Box.createHorizontalGlue());
	    mb.add(mHelp);
	  }
	  // MAIN 0
	  setTitle(TGS.lt[0]);
	  setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
	  {
	    setJMenuBar(mb);
	    Container cp = null;
	    cp = getContentPane();
	    cp.add(tbButtons, BorderLayout.PAGE_START);
	    cp.add(pContents, BorderLayout.CENTER);
	  }
	  bDirectory.addActionListener(this);
	  bExit.addActionListener(this);
	  bOptions.addActionListener(this);
	  bRun.addActionListener(this);
	  lCurrentDir.setTransferHandler(dh);
	  lCurrentStatus.setTransferHandler(dh);
	  lDir.setTransferHandler(dh);
	  lIcon.setTransferHandler(dh);
	  lStatus.setTransferHandler(dh);
	  mAbout.addActionListener(this);
	  mDirectory.addActionListener(this);
	  mExit.addActionListener(this);
	  mGui.addActionListener(this);
	  mLocalHelp.addActionListener(this);
	  mOptions.addActionListener(this);
	  mRun.addActionListener(this);
	  mWebHelp.addActionListener(this);
	  pContents.setTransferHandler(dh);
	}
// END OF JLAYOUT SECTION
// jlayout.setup.end

// jlayout.cleanup.begin
// START OF JLAYOUT SECTION
// This section was generated automatically by the jlayout program.
// You should not modify this section.
// Any changes applied to this section will be lost if jlayout is run again.
	private void jLayoutCleanup()
	{
	  bDirectory.removeActionListener(this);
	  bExit.removeActionListener(this);
	  bOptions.removeActionListener(this);
	  bRun.removeActionListener(this);
	  mAbout.removeActionListener(this);
	  mDirectory.removeActionListener(this);
	  mExit.removeActionListener(this);
	  mGui.removeActionListener(this);
	  mLocalHelp.removeActionListener(this);
	  mOptions.removeActionListener(this);
	  mRun.removeActionListener(this);
	  mWebHelp.removeActionListener(this);
	}
// END OF JLAYOUT SECTION
// jlayout.cleanup.end

  public TraceccMainWindow(TraceccController t, GuiController g) {
    super(TGS.lt[0], g);
    // save controllers passed as arguments
    tc = t; gc = g;
    // prepare variables needed by jLayoutSetup
    b = gc.getBorder();
    cGreen = new Color(0, 128, 0); cRed = new Color(128, 0, 0);
    dh = new DropHandler();
    // run jLayoutSetup
    jLayoutSetup();
    // set label for current directory, keep track of size
    lCurrentDir.setText(getShortDirectoryName(tc.getString(0)));
    lCurrentDir.setToolTipText(tc.getString(0));
    stDirectory = new SizeTracker();
    stDirectory.mustPackAgain(lCurrentDir);
    stStatus = new SizeTracker();
    stStatus.mustPackAgain(lStatus);
    stCurrentStatus = new SizeTracker();
    stCurrentStatus.mustPackAgain(lCurrentStatus);
    pack();
  }

  @Override
  public void cleanup(boolean isLast) {
    TraceccOptionsDialog t = getTod();
    if(t != null) {
        t.setModal(false); t.setVisible(false);
        setTod(null); t = null;
    }
    GuiDialog g = getGuid();
    if(g != null) {
        g.setModal(false); g.setVisible(false);
        setGuid(null); g = null;
    }
    if(isLast) {
      tc.save();
    }
    jLayoutCleanup();
    gc = null;
    tc = null;
  }



  /**
   * Enable/disable GUI components.
   *
   * @param b
   * Flag to enable.
   */
  private void setGuiEnabled(boolean b) {
    mDirectory.setEnabled(b);
    mOptions.setEnabled(b);
    mRun.setEnabled(b);
    mExit.setEnabled(b);
    bDirectory.setEnabled(b);
    bOptions.setEnabled(b);
    bRun.setEnabled(b);
    bExit.setEnabled(b);
  }



  public String getShortDirectoryName(String s) {
    String back = s;
    int i = s.lastIndexOf(File.separator);
    if(i >= 0) {
      i += File.separator.length();
      StringBuilder sb = new StringBuilder();
      sb.append("...");
      sb.append(File.separator);
      sb.append(s.substring(i));
      back = sb.toString();
    }
    return back;
  }



  private void setNewDirectory(String s) {
    tc.setString(0, s);
    String t = getShortDirectoryName(s);
    lCurrentDir.setText(t);
    lCurrentDir.setToolTipText(s);
    if(stDirectory.mustPackAgain(lCurrentDir)) {
      pack();
      ComponentPositioner.correctPosition(this);
    }
    repaint();
  }



  private void chooseDirectory() {
    JFileChooser c = new JFileChooser(tc.getString(0));
    c.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
    c.setFileFilter(new DirectoryFilter());
    int result = c.showOpenDialog(this);
    if(result == JFileChooser.APPROVE_OPTION) {
      String wd = c.getSelectedFile().getAbsolutePath();
      setNewDirectory(wd);
    }
  }



  private void showOptionsDialog() {
    TraceccOptionsDialog toptd = useTod();
    toptd.controller_to_dialog();
    toptd.setModal(true);
    toptd.setVisible(true);
  }



  private void disableGui() {
      dh.setEnabled(false);
  }



  private void enableGui() {
    dh.setEnabled(true);
    bDirectory.setEnabled(true);
    bExit.setEnabled(true);
    bOptions.setEnabled(true);
    bRun.setEnabled(true);
    mDirectory.setEnabled(true);
    mOptions.setEnabled(true);
    mRun.setEnabled(true);
    mExit.setEnabled(true);
  }


  private int toRange(int v, int min, int max) {
    int back = v;
    if(back < min) { back = min; }
    if(back > max) { back = max; }
    return back;
  }



  private void handleListOfFiles(
    TraceccController tc,
    LogContentsForCommand lcfc,
    HashSet<String> hs,
    String s1,
    String s2
  ) {
    int i, fnl, lios;
    int s1l = s1.length();
    Object[] objarray = hs.toArray();
    for(i = 0; i < objarray.length; i++) {
      String filename = (String)objarray[i];
      fnl = filename.length();
      if(fnl > s1l) {
        StringBuilder sb = new StringBuilder();
        sb.append(filename.substring(0, (fnl - s1l)));
        sb.append(s2);
        String outputname = sb.toString();
        lios = filename.lastIndexOf(File.separator);
	if(lios >= 0) {
          lios += File.separator.length();
	} else {
	  lios = 0;
	}
        SwingUtilities.invokeLater(
          new StatusChange(TGS.lt[45], filename.substring(lios), cRed)
        );
        // try { Thread.sleep(200); } catch(Exception exc) {}
        lcfc.newCommand();
        lcfc.setDirectory(tc.getString(0));
        lcfc.addArgv(TraceccController.pao[0]);
        lcfc.addArgv(TraceccController.pao[1]);
        if(tc.bv[7]) {      // line numbers
          lcfc.addArgv(TraceccController.pao[2]);
        }
        if(tc.bv[8]) {      // C++ comments
          lcfc.addArgv(TraceccController.pao[3]);
        }
        lcfc.addArgv(TraceccController.pao[4]);
        lcfc.addArgv((new Integer(toRange(tc.iv[0], 30, 127))).toString());
        if(tc.bv[9]) {      // debug
          lcfc.addArgv(TraceccController.pao[5]);
          if(tc.bv[10]) {   // debug stdout
            lcfc.addArgv(TraceccController.pao[6]);
          }
          if(tc.bv[11]) {   // debug IDE
            lcfc.addArgv(TraceccController.pao[7]);
            lcfc.addArgv(
              TGS.idedef[
                toRange(tc.iv[1], 0, (TGS.idedef.length - 1))
              ]
            );
          }
          if(tc.bv[12]) {   // time stamp for debug
            lcfc.addArgv(TraceccController.pao[9]);
          }
          if(tc.bv[13]) {   // trace keyword
            lcfc.addArgv(TraceccController.pao[8]);
          }

        }
        lcfc.addArgv(TraceccController.pao[10]);
        lcfc.addArgv(TraceccController.pao[11]);
        lcfc.addArgv(TraceccController.pao[12]);
        lcfc.addArgv(TraceccController.pao[13]);
        lcfc.addArgv(TraceccController.pao[14]);
        lcfc.addArgv(filename);
        lcfc.addArgv(outputname);
        lcfc.run();
      }
    }
  }

  private void handleJlayout(
    TraceccController mytc,
    LogContentsForCommand lcfc,
    TraceccListOfInputFiles tloif
  ) {
    int i, s1l, fnl, lios;
    String filename;
    Object[] objarray = tloif.jlFiles.toArray();
    s1l = TraceccListOfInputFiles.suffixes[8].length();
    for(i = 0; i < objarray.length; i++) {
      filename = (String)objarray[i];
      fnl = filename.length();
      if(fnl > s1l) {
        StringBuilder sb = new StringBuilder();
	sb.append(filename.substring(0, (fnl - s1l)));
	sb.append(TraceccListOfInputFiles.suffixes[7]);
	String outputname = sb.toString();
	File of = new File(outputname);
	if(of.exists()) {
	  if(of.isFile()) {
	    lios = filename.lastIndexOf(File.separator);
	    if(lios >= 0) {
	      lios += File.separator.length();
	    } else {
	      lios = 0;
	    }
            SwingUtilities.invokeLater(
              new StatusChange(TGS.lt[46], filename.substring(lios), cRed)
            );
            // try { Thread.sleep(200); } catch(Exception exc) {}
	    lcfc.newCommand();
	    lcfc.setDirectory(mytc.getString(0));
	    lcfc.addArgv(TraceccController.pao[15]);
	    for(lios = 10; lios <= 14; lios++) {
	      lcfc.addArgv(TraceccController.pao[lios]);
	    }
	    lcfc.addArgv(filename);
	    lcfc.addArgv(outputname);
	    lcfc.run();
	  }
	}
      }
    }
  }

  private void backgroundThreadFunction() {
      TraceccController mytc = new TraceccController(null);
      mytc.cloneFrom(tc);
      SwingUtilities.invokeLater(
        new StatusChange(TGS.lt[44], TGS.lt[11], cRed)
      );
      TraceccListOfInputFiles tloif = new TraceccListOfInputFiles(mytc);
      tloif.getFilenames();
      LogContentsForCommand lcfc = new LogContentsForCommand();
      lcfc.setShowAlways(mytc.bv[0]);
      handleListOfFiles(
        mytc, lcfc, tloif.ctrFiles,
        TraceccListOfInputFiles.suffixes[0],
        TraceccListOfInputFiles.suffixes[1]
      );
      handleListOfFiles(
        mytc, lcfc, tloif.cptFiles,
        TraceccListOfInputFiles.suffixes[2],
        TraceccListOfInputFiles.suffixes[3]
      );
      handleListOfFiles(
        mytc, lcfc, tloif.mtrFiles,
        TraceccListOfInputFiles.suffixes[4],
        TraceccListOfInputFiles.suffixes[5]
      );
      handleListOfFiles(
        mytc, lcfc, tloif.jtrFiles,
        TraceccListOfInputFiles.suffixes[6],
        TraceccListOfInputFiles.suffixes[7]
      );
      // TODO: show "jlayout"
      handleJlayout(mytc, lcfc, tloif);
      // Enable all the GUI components
      // in AWT-Thread: dh.setEnabled(true);
      setRunning(false, null);
      SwingUtilities.invokeLater(new ConversionDone(mytc, lcfc));
  }

  private void runBackgroundThread() {
    Thread t = new Thread(new BackgroundThread());
    setRunning(true, t);
    dh.setEnabled(false);
    bDirectory.setEnabled(false);
    bExit.setEnabled(false);
    bOptions.setEnabled(false);
    bRun.setEnabled(false);
    mDirectory.setEnabled(false);
    mOptions.setEnabled(false);
    mRun.setEnabled(false);
    mExit.setEnabled(false);
    t.start();
  }



  private void showGuiDialog() {
    GuiDialog g = useGuid();
    g.controllerToDialog();
    g.setModal(true);
    g.setVisible(true);
  }



  private void conversionDone(TraceccController t, LogContentsForCommand lcfc) {
    lStatus.setText(TGS.lt[9]);
    lStatus.setForeground(cGreen);
    lCurrentStatus.setText(TGS.lt[11]);
    enableGui();
    repaint();
    boolean mustShowLog = true;
    if(lcfc.getSuccess()) { mustShowLog = false; }
    if(t.bv[0]) { mustShowLog = true; }
    if(mustShowLog) {
      LogWindow lw = useLogWindow();
      lw.setTitle(TGS.lt[43]);
      lw.showTextIfNecessary(lcfc);
      lw.setVisible(true);
    }
  }



  private void changeStatus(String s1, String s2, Color co) {
    lStatus.setText(s1); lCurrentStatus.setText(s2);
    lStatus.setForeground(co);
    boolean mustPackAgain = stStatus.mustPackAgain(lStatus);
    if(stCurrentStatus.mustPackAgain(lCurrentStatus)) {
      mustPackAgain = true;
    }
    if(mustPackAgain) {
      pack();
      ComponentPositioner.correctPosition(this);
    }
    repaint();
  }



  private void showLocalHelp() {
    URL helpUrl = null;
        HelpWindow lwn = useHelpWindow();
	lwn.setPurpose(1);
	lwn.setContentType(0);
	lwn.setTitle(TGS.lt[41]);
        helpUrl = TraceccMainWindow.class.getResource(TGS.lt[42]);
        if(helpUrl != null) {
          try {
            lwn.setPage(helpUrl);
          }
          catch(IOException ioe) {
            lwn.setText(LogWindow.theTexts[1]);
          }
        } else {
        }
	if(lwn.isVisible()) {
	  lwn.repaint();
	} else {
	  lwn.setVisible(true);
	}
  }



  private void launchWebHelp() {
    try {
      Desktop.getDesktop().browse(new URI(TGS.projectUri));
    } catch(Exception exc) {}
  }



  private void showAboutDialog() {
        HelpWindow lwn = useHelpWindow();
	lwn.setPurpose(2);
	lwn.setContentType(1);
	lwn.setTitle(TGS.lt[40]);
	lwn.setText(TGS.licenseTerms);
	if(lwn.isVisible()) {
	  lwn.repaint();
	} else {
	  lwn.setVisible(true);
	}
  }



  private void exitApplication() {
    processWindowEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
  }



  @Override
  public void actionPerformed(ActionEvent e) {
    switch(StringTool.getArrayIndex(TGS.ac, e.getActionCommand())) {
      case 0: {	// directory
        if(!isRunning()) {
	  chooseDirectory();
	}
      } break;
      case 1: {	// options
        if(!isRunning()) {
	  showOptionsDialog();
	}
      } break;
      case 2: {	// run
        if(!isRunning()) {
	  runBackgroundThread();
	}
      } break;
      case 3: {	// exit
        if(!isRunning()) {
	  exitApplication();
	}
      } break;
      case 4: {	// GUI
        showGuiDialog();
      } break;
      case 5: {	// local help
        showLocalHelp();
      } break;
      case 6: {	// web site
        launchWebHelp();
      } break;
      case 7: {	// about
        showAboutDialog();
      } break;
    }
  }



}



// project tracegui
// @(#)TraceccMainWindow.java 1.2 07/09/08

