

#include <stdio.h>
#include <process.h>
#include <io.h>




#line 9 "dkwftreg.ctr"



#define CMD_BMEPS_FILE		1
#define CMD_FIG2VECT_FILE	2
#define CMD_TRACECC_DIR		4
#define CMD_BMEPS_DIR		8
#define CMD_FIG2VECT_DIR	16


#include "dkstr.h"
#include "dksf.h"

#include <windows.h>


#define LINESIZE 1024



/* Return value of the main function */
static int exval = 0;



static int install_cmd = (
  CMD_BMEPS_FILE
  | CMD_BMEPS_DIR
  | CMD_TRACECC_DIR
  | CMD_FIG2VECT_FILE
);



/** Data to register context menu entries. */
typedef struct {
  char  *tt;  /**< Title text. */
  char  *cm;  /**< Command to run. */
  char	*ar;  /**< Command arguments. */
} FileAction;



static char default_base_dir[] = { "C:\\Programme\\Krause" };
static char buffer1[1024];
static char buffer2[1024];
static char part1[] = { "-jar " };
static char part2[] = { "\\bin\\bmepsgui.jar" };
static char part3[] = { "\\bin\\tracegui.jar" };
static char fn1[] = { "\"%1\"" };
static char str_sp[] = { " " };
static char str_dq[] = { "\"" };

/*
  These settings are applied to .png, .jpg, .jpeg, .pbm, .pgm, .ppm, .pnm, .tif and .tiff files.
*/
FileAction bmeps_actions[] = {
  {
    "bmeps: Convert to BoundingBox",
    "bmeps.exe",
    "-lbb -a \"%1\""
  },
  {
    "bmeps: Convert to EPS1",
    "bmeps.exe",
    "-leps1 -a \"%1\""
  },
  {
    "bmeps: Convert to EPS2",
    "bmeps.exe",
    "-leps2 -a \"%1\""
  },
  {
    "bmeps: Convert to EPS3",
    "bmeps.exe",
    "-leps3 -a \"%1\""
  },
  {
    "bmeps: Convert to PDF",
    "bmeps.exe",
    "-lpdf -a -A \"%1\""
  },

  { NULL, NULL, NULL }
};




/*
  These settings are applied to .tif and .tiff files.
*/
FileAction bmeps_fax_actions[] = {
  {
    "bmeps: Fax to PDF",
    "bmeps.exe",
    "-lfaxpdf -a -A \"%1\""
  },
  { NULL, NULL, NULL }
};



/*
  These settings are applied to directories.
*/
FileAction tracecc_actions[] = {
  {
    "Open in Tracecc GUI",
    "javaw.exe",
    NULL
  },
  { NULL, NULL, NULL }
};



/*
	These actions are applied to directories.
*/
FileAction bmeps_dir_actions[] = {
  {
    "Open in Bmeps GUI",
    "javaw.exe",
    NULL
  },
  { NULL, NULL, NULL }
};



FileAction fig2vect_actions[] = {
  {
    "fig2vect: Convert to EPS",
    "fig2vect.exe",
    "-leps -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to PDF",
    "fig2vect.exe",
    "-lpdf -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to EPS.TEX",
    "fig2vect.exe",
    "-leps.tex -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to PDF.TEX",
    "fig2vect.exe",
    "-lpdf.tex -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to TEX",
    "fig2vect.exe",
    "-ltex -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to TEX FULL",
    "fig2vect.exe",
    "-ltex.full -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to BB",
    "fig2vect.exe",
    "-lbb -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to SVG",
    "fig2vect.exe",
    "-lsvg -a --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  { NULL, NULL, NULL}
};



FileAction fig2vect_directory_actions[] = {
  {
    "fig2vect: Convert to EPS",
    "fig2vect.exe",
    "-leps -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to PDF",
    "fig2vect.exe",
    "-lpdf -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to EPS.TEX",
    "fig2vect.exe",
    "-leps.tex -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to PDF.TEX",
    "fig2vect.exe",
    "-lpdf.tex -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to TEX",
    "fig2vect.exe",
    "-ltex -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to EPS FULL",
    "fig2vect.exe",
    "-ltex.full -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to BB",
    "fig2vect.exe",
    "-lbb -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to SVG",
    "fig2vect.exe",
    "-lsvg -a -m- --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to EPS (make)",
    "fig2vect.exe",
    "-leps -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to PDF (make)",
    "fig2vect.exe",
    "-lpdf -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to EPS.TEX (make)",
    "fig2vect.exe",
    "-leps.tex -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to PDF.TEX (make)",
    "fig2vect.exe",
    "-lpdf.tex -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to TEX (make)",
    "fig2vect.exe",
    "-ltex -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to EPS FULL (make)",
    "fig2vect.exe",
    "-ltex.full -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to BB (make)",
    "fig2vect.exe",
    "-lbb -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  {
    "fig2vect: Convert to SVG (make)",
    "fig2vect.exe",
    "-lsvg -a -m --/log/stdout/level=none --/log/stderr/level=progress \"%1\""
  },
  { NULL, NULL, NULL}
};


/** Subdirectory bin. */
static char subdirectory[] = { "\\bin" };

/** Name of environment variable to change. */
static char str_path[] = { "PATH" };

static char str_key[] =
{ "System\\CurrentControlSet\\Control\\Session Manager\\Environment" };

static char str_semicolon[] = { ";" };

static
int
get_environment_variable(char *b, size_t s, char *n) {
  int back = 0;
  LONG result;
  HKEY base = HKEY_LOCAL_MACHINE;
  HKEY reskey;
  DWORD dwType, dwSz;
  
  result = RegCreateKeyExA(
    base, str_key, (DWORD)0, NULL, REG_OPTION_NON_VOLATILE,
    KEY_ALL_ACCESS, NULL, &reskey, NULL
  );
  if(result == ERROR_SUCCESS) {	
    dwType = REG_SZ; dwSz = s;
    result = RegQueryValueExA(reskey, n, NULL, &dwType, (BYTE *)b, &dwSz);
    if(result == ERROR_SUCCESS) {	
      if((dwType == REG_SZ) || (REG_EXPAND_SZ)) {		
	if(dwSz > 0) {			
	  b[(dwSz < s) ? dwSz : (s-1)] = '\0';
	  back = 1;
	  
	} else {			
	}
      } else {				
        if(dwType == REG_EXPAND_SZ) {	
	}
      }
    } else {				
    }
    RegCloseKey(reskey);
  } else {	
  }
  
  return back;
}



static
int
set_environment_variable(char *k, char *v) {
  int back = 0;
  char oldval[4096];
  LONG result;
  HKEY base = HKEY_LOCAL_MACHINE;
  HKEY reskey;
  DWORD newDwType, dwType, dwSz;
  
  result = RegCreateKeyExA(
    base, str_key, (DWORD)0, NULL, REG_OPTION_NON_VOLATILE,
    KEY_ALL_ACCESS, NULL, &reskey, NULL
  );
  if(result == ERROR_SUCCESS) {		
    newDwType = REG_SZ;
    dwType = REG_SZ; dwSz = (DWORD)sizeof(oldval);
    result = RegQueryValueExA(reskey, k, NULL, &dwType, (BYTE *)oldval, &dwSz);
    if(result == ERROR_SUCCESS) {	
      newDwType = dwType;
    }
    if((newDwType == REG_SZ) || (newDwType == REG_EXPAND_SZ)) {
      
      result = RegSetValueExA(
        reskey, k, (DWORD)0, newDwType, (BYTE *)v, (DWORD)(1+strlen(v))
      );
      if(result == ERROR_SUCCESS) { back = 1; }
    }
    RegCloseKey(reskey);
  } else {				
  } 
  return back;
}



static
int
path_contains(char *fullpath, char *dir) {
  int back = 0;
  char *p1, *p2;
  p1 = fullpath;
  while((p1) && (back == 0)) {
    p2 = strchr(p1, ';');
    if(p2) {
      *p2 = '\0';
    }
    if(stricmp(p1, dir) == 0) {
      back = 1;
    }
    if(p2) {
      *p2 = ';'; p2++;
    }
    p1 = p2;
  }
  return back;
}



static
void
register_directory_in_path_if_necessary(char *s) {
  char cpldir[4096];	/**< Complete directory name. */
  char oldval[4096];	/**< Current PATH environment variable. */
  char newval[4096];	/**< New PATH environment variable. */
  
  if((strlen(s) + strlen(subdirectory)) < sizeof(cpldir)) {
    strcpy(cpldir, s);
    strcat(cpldir, subdirectory);
    if(get_environment_variable(oldval, sizeof(oldval), str_path)) {
      
      if(!path_contains(oldval, cpldir)) {
        
        if((strlen(cpldir) + strlen(oldval) + 2) < sizeof(newval)) {
	  strcpy(newval, cpldir);
	  strcat(newval, str_semicolon);
	  strcat(newval, oldval);
	  
	  set_environment_variable(str_path, newval);
	}
      }
    }
  }
  
}



static
void
unregister_directory_in_path_if_necessary(char *s) {
  char cpldir[4096];	/**< Complete directory name. */
  char oldval[4096];	/**< Current PATH environment variable. */
  char newval[4096];	/**< New PATH environment variable. */
  int  values_found;	/**< Flag: this is not the first directory. */
  char *p1;		/**< Pointer to current directory. */
  char *p2;		/**< Pointer to next directory to process. */
  
  if((strlen(s) + strlen(subdirectory)) < sizeof(cpldir)) {
    strcpy(cpldir, s);
    strcat(cpldir, subdirectory);
    if(get_environment_variable(oldval, sizeof(oldval), str_path)) {
      p1 = oldval; values_found = 0;
      newval[0] = '\0';
      while(p1) {
        p2 = strchr(p1, ';');
	if(p2) { *p2 = '\0'; }
	
	if(stricmp(p1, cpldir)) {
	  if(values_found) {
	    strcat(newval, str_semicolon);
	    strcat(newval, p1);
	  } else {
	    strcpy(newval, p1);
	  }
	  values_found = 1;
	}
	if(p2) { *(p2++) = ';'; }
	p1 = p2;
      } 
      set_environment_variable(str_path, newval);
    }
  }
  
}



/*
	Open a registry key.
	Arguments:
	  base		Base key.
	  n		Name of subkey.
	  reskey	Pointer to HKEY to store the result.
	Returns:
	  1 ... Success
	  0 ... Error
*/
static
int
open_registry_key(HKEY base, char *n, HKEY *reskey) {
  int back = 0;
  LONG result;
  
  result = RegCreateKeyExA(
    base, n, (DWORD)0, NULL, REG_OPTION_NON_VOLATILE,
    KEY_ALL_ACCESS, NULL, reskey, NULL
  );
  if(result == ERROR_SUCCESS) {
    back = 1;
  } 
  return back;
}



/*
	Set string registry entry.
	Parameters:
	  hk	Key to create the entry
	  n	The name of the entry
	  n	Entry value
	Returns:
	  1 ... Success
	  0 ... Error
*/
static
int
set_string_entry(HKEY hk, char *n, char *v) {
  int back = 0;
  long result;
  
  result = RegSetValueExA(
    hk, n, (DWORD)0, REG_SZ, (BYTE *)v, (DWORD)(1 + strlen(v))
  );
  if(result == ERROR_SUCCESS) { back = 1; } 
  return back;
}



/*
	Retrieve string registry entry.
	Parameters:
	  hk	The key to retrieve the entry from
	  n	Name of the entry
	  b	Buffer to store result
	  sz	Result buffer size
	Returns:
	  1 ... Success
	  0 ... Error
*/
static
int
get_string_entry(HKEY hk, char *n, char *b, size_t sz) {
  int back = 0;
  long result;
  DWORD dwSz, dwType;
  dwSz = sz; dwType = REG_SZ;
  
  result = RegQueryValueExA(
    hk, n, NULL, &dwType, (BYTE *)b, &dwSz
  ); 
  if(result == ERROR_SUCCESS) {	
    if(dwType == REG_SZ) {
    if(dwSz > 0) {	
      b[(dwSz < sz) ? dwSz : (sz - 1)] = '\0';
      back = 1;		
    }
    }
  } else {	
  } 
  return back;
}


static char str_bs[] = { "\\" };
static char str_shell[] = { "shell" };
static char str_command[] = { "command" };
static char str_bin[] = { "bin" };



/*
	Remove a context menu entry.
	Parameters:
	  act	Array listing all menu entries
	  ft	File type to remove actions from
	  dn	Directory where bmeps and tracecc reside
*/
static
void
unregister_filetype_action(FileAction *act, char *ft, char *dn) {
  char keyname[LINESIZE];
  size_t sz;
  FileAction *actptr;
  
  actptr = act;

  
  
  
  while((actptr->tt) || (actptr->cm) || (actptr->ar)) {
    if(actptr->tt) {
      
      
      
      sz = strlen(ft);
      sz += 2 * strlen(str_bs);
      sz += strlen(str_shell);
      sz += strlen(actptr->tt);
      if(sz < sizeof(keyname)) {	
        sz += strlen(str_bs); sz += strlen(str_command);
	if(sz < sizeof(keyname)) {	
	  strcpy(keyname, ft);
	  strcat(keyname, str_bs);
	  strcat(keyname, str_shell);
	  strcat(keyname, str_bs);
	  strcat(keyname, actptr->tt);
	  strcat(keyname, str_bs);
	  strcat(keyname, str_command);
	  
	  RegDeleteKeyA(HKEY_CLASSES_ROOT, keyname);
	} else {			
	}
	strcpy(keyname, ft);
	strcat(keyname, str_bs);
	strcat(keyname, str_shell);
	strcat(keyname, str_bs);
	strcat(keyname, actptr->tt);
	
	RegDeleteKeyA(HKEY_CLASSES_ROOT, keyname);
      } else {		
      }
    }
    actptr++;
  }
  
}



/*	
	Remove context menu entries for file name suffix.
	Parameters:
	  fn	File name suffix (i.e. ".png")
	  act	List of context menu entries to remove
	  ft	File type default
	  dn	Directory name where bmeps and tracecc reside
*/
static
void
unregister_action(char *fns, FileAction *act, char *ft, char *dn) { 
  HKEY hk;
  char filetypebuffer[LINESIZE]; int have_filetype = 0;
  
  if(open_registry_key(HKEY_CLASSES_ROOT, fns, &hk)) {
    
    if(get_string_entry(hk, "", filetypebuffer, sizeof(filetypebuffer))) {
      have_filetype = 1;	
    } else {			
      if(strlen(ft) < sizeof(filetypebuffer)) {	
        strcpy(filetypebuffer, ft); have_filetype = 1;
      }
    }	
    RegCloseKey(hk);	
    if(have_filetype) {
      if(open_registry_key(HKEY_CLASSES_ROOT, filetypebuffer, &hk)) {
        RegCloseKey(hk);
	unregister_filetype_action(act, filetypebuffer, dn);
      }
    }
  } else {		
    /* ##### FAILED TO OPEN/CREATE KEY */
  }
  unregister_filetype_action(act, ft, dn);
  
}



/*
	Register context menu entries for file type.
	Parameters:
	  act	List of context menu entries
	  ft	File type
	  dn	Directory where tracecc and bmeps reside
*/
static
void
register_filetype_action(FileAction *act, char *ft, char *dn) {
  HKEY hk;
  char keyname[LINESIZE];
  char valuebuffer[LINESIZE];
  FileAction *actptr;
  size_t sz;
  sz = strlen(ft);
  sz += strlen(str_bs); sz += strlen(str_shell);
  if(sz < sizeof(keyname)) {
    strcpy(keyname, ft);
    strcat(keyname, str_bs); strcat(keyname, str_shell);
    if(open_registry_key(HKEY_CLASSES_ROOT, keyname, &hk)) {
      RegCloseKey(hk);
      actptr = act;
      while((actptr->tt) || (actptr->cm) || (actptr->ar)) {
        if((actptr->tt) && (actptr->cm) && (actptr->ar)) {
          sz = strlen(ft);
  	  sz += strlen(str_bs); sz += strlen(str_shell);
          sz += strlen(str_bs); sz += strlen(actptr->tt);
  	  if(sz < sizeof(keyname)) {
  	    strcpy(keyname, ft);
  	    strcat(keyname, str_bs); strcat(keyname, str_shell);
  	    strcat(keyname, str_bs); strcat(keyname, actptr->tt);
  	    if(open_registry_key(HKEY_CLASSES_ROOT, keyname, &hk)) {
  	      RegCloseKey(hk);
  	      sz += strlen(str_bs); sz += strlen(str_command);
  	      if(sz < sizeof(keyname)) {
  	        strcat(keyname, str_bs); strcat(keyname, str_command);
  	        if(open_registry_key(HKEY_CLASSES_ROOT, keyname, &hk)) {
  	          sz = strlen(actptr->cm);
  		  sz += 2 * strlen(str_dq);
  		  if(dn) {
  		    sz += strlen(dn); sz += 2 * strlen(str_bs);
  		    sz += strlen(str_bin);
  		  }
  		  sz += strlen(actptr->ar);
  		  sz += 1;
  		  if(sz < sizeof(valuebuffer)) {
  		    strcpy(valuebuffer, str_dq);
  		    if(dn) {
  		      strcat(valuebuffer, dn);
  		      strcat(valuebuffer, str_bs);
  		      strcat(valuebuffer, str_bin);
  		      strcat(valuebuffer, str_bs);
  		    }
  		    strcat(valuebuffer, actptr->cm);
  		    strcat(valuebuffer, str_dq);
  		    strcat(valuebuffer, str_sp);
  		    strcat(valuebuffer, actptr->ar);
  		    set_string_entry(hk, "", valuebuffer);
  		  }
  		  RegCloseKey(hk);
  	        }
  	      }
  	    }
  	  }
        }
        actptr++;
      }
    }
  }
}



/*
	Register context menu entries for file name suffix.
	Parameters
	  fns	File name suffix (i.e. ".png")
	  act	List of context menu entries
	  ft	Default file type name
	  dn	Directory where bmeps and tracecc reside
*/
static
void
register_action(char *fns, FileAction *act, char *ft, char *dn) {
  HKEY hk;
  char filetypebuffer[LINESIZE]; int have_filetype = 0;
  
  if(open_registry_key(HKEY_CLASSES_ROOT, fns, &hk)) {
    
    if(get_string_entry(hk, "", filetypebuffer, sizeof(filetypebuffer))) {
      have_filetype = 1;	
    } else {			
      if(strlen(ft) < sizeof(filetypebuffer)) {	
        strcpy(filetypebuffer, ft); have_filetype = 1;
      }
    }	
    RegCloseKey(hk);	
    if(have_filetype) {
      if(open_registry_key(HKEY_CLASSES_ROOT, filetypebuffer, &hk)) {
        RegCloseKey(hk);
	register_filetype_action(act, filetypebuffer, dn);
      }
    }
  } else {		
    /* ##### FAILED TO OPEN/CREATE KEY */
  } 
}



/*	
	DLLEXPORT Create context menu entries for bitmap files
	to invoke bmeps.
*/
void
BmepsContextMenuEntries(char *dirname) {
  register_action(".png",  bmeps_actions, "pngfile", dirname);
  register_action(".jpg",  bmeps_actions, "jpegfile", dirname);
  register_action(".jpeg", bmeps_actions, "jpegfile", dirname);
  register_action(".pgm",  bmeps_actions, "netpbmfile", dirname);
  register_action(".pbm",  bmeps_actions, "netpbmfile", dirname);
  register_action(".ppm",  bmeps_actions, "netpbmfile", dirname);
  register_action(".pnm",  bmeps_actions, "netpbmfile", dirname);
  register_action(".tif",  bmeps_actions, "tifffile", dirname);
  register_action(".tiff", bmeps_actions, "tifffile", dirname);
  register_action(".tif",  bmeps_fax_actions, "tifffile", dirname);
  register_action(".tiff", bmeps_fax_actions, "tifffile", dirname);
}



/*
	DLLEXPORT Create context menu entries for directories
	to invoke bmeps
*/
void
BmepsDirectoryContextMenuEntries(char *dirname) {
  register_filetype_action(bmeps_dir_actions, "Directory", NULL);
}



/*
	DLLEXPORT Create context menu entries for directories
	to invoke tracecc.
*/
void
TraceccDirectoryContextMenuEntries(char *dirname) {
  register_filetype_action(tracecc_actions, "Directory", NULL);
}


void
FigContextMenuEntries(char *dirname) {
  register_action(".fig", fig2vect_actions, "figfile", dirname);
}


void
FigDirectoryContextMenuEntries(char *dirname) {
  register_filetype_action(fig2vect_directory_actions, "Directory", dirname);
}


/*
	Create all context menu entries
*/
static
void
register_actions(char *dirname) {
  
  if(install_cmd & CMD_BMEPS_FILE) {
    BmepsContextMenuEntries(dirname);
  }
  if(install_cmd & CMD_BMEPS_DIR) {
    BmepsDirectoryContextMenuEntries(dirname);
  }
  if(install_cmd & CMD_TRACECC_DIR) {
    TraceccDirectoryContextMenuEntries(dirname);
  }
  if(install_cmd & CMD_FIG2VECT_FILE) {
    FigContextMenuEntries(dirname);
  }
  if(install_cmd & CMD_FIG2VECT_DIR) {
    FigDirectoryContextMenuEntries(dirname);
  }
  
}



/*
	DLLEXPORT Remove all context menu entries
*/
void
RemoveContextMenuEntries(char *dirname) {
  
  unregister_action(".png",  bmeps_actions, "pngfile", dirname);
  unregister_action(".jpg",  bmeps_actions, "jpegfile", dirname);
  unregister_action(".jpeg", bmeps_actions, "jpegfile", dirname);
  unregister_action(".pgm",  bmeps_actions, "netpbmfile", dirname);
  unregister_action(".pbm",  bmeps_actions, "netpbmfile", dirname);
  unregister_action(".ppm",  bmeps_actions, "netpbmfile", dirname);
  unregister_action(".pnm",  bmeps_actions, "netpbmfile", dirname);
  unregister_action(".tif",  bmeps_actions, "tifffile", dirname);
  unregister_action(".tiff", bmeps_actions, "tifffile", dirname);
  unregister_action(".tif",  bmeps_fax_actions, "tifffile", dirname);
  unregister_action(".tiff", bmeps_fax_actions, "tifffile", dirname);
  unregister_action(".fig", fig2vect_actions, "figfile", dirname);
  unregister_filetype_action(bmeps_dir_actions, "Directory", NULL);
  unregister_filetype_action(tracecc_actions, "Directory", NULL);
  unregister_filetype_action(fig2vect_directory_actions, "Directory", dirname);
  
}



static char *q1[] = {
"Do you wish to install or uninstall context menu entries?",
"1   Install",
"2   Uninstall",
"q   Quit",
NULL
};


static
int
install_or_uninstall(void) {
  char  buffer[1024];
  int back = 0, value = 0;  char **ptr;
  ptr = q1; while(*ptr) { printf("%s\n", *(ptr++)); }
  if(fgets(buffer, sizeof(buffer), stdin)) {
    if(sscanf(buffer, "%d", &value) == 1) {
      switch(value) {
        case 1: back = 0; break;
        case 2: back = 1; break;
        default: back = -1; break;
      }
    } else {
      back = -1;
    }
  } else {
    back = -1;
  }
  return back;
}

static char *q2[] = {
"Add bmeps context menu entries to bitmap files? (Y/n): ",
"Add fig2vect context menu entries to *.fig files? (Y/n): ",
"Add tracecc context menu entries to directories? (Y/n): ",
"Add bmeps context menu entries to directories (Y/N): ",
"Add fig2vect context menu entries to directories? (y/N): ",
NULL
};


static
int
mod_install_cmd(int i, int v, int d) {
  char buffer[256], *p1;
  int back = 0;
  int m = d;
  printf("%s", q2[i]);
  if(fgets(buffer, sizeof(buffer), stdin)) {
    p1 = dkstr_start(buffer, NULL);
    if(p1) {
      dkstr_chomp(p1, NULL);
      if(dkstr_is_bool(p1)) {
        back = (dkstr_is_on(p1) ? v : 0);
      }
    }
  }
  if(m) {
    back = v;
  }
  return back;
}



static
void
setup_buffers(char *dn) {
  size_t sz;
  if(dn) {
    sz = strlen(part1) + strlen(dn) + strlen(part2) + strlen(fn1) + strlen(str_sp) + 3;
    if(sz < sizeof(buffer1)) {
      strcpy(buffer1, part1);
      strcat(buffer1, str_dq);
      strcat(buffer1, dn);
      strcat(buffer1, part2);
      strcat(buffer1, str_dq);
      strcat(buffer1, str_sp);
      strcat(buffer1, fn1);
      bmeps_dir_actions[0].ar = buffer1;
    }
    sz = strlen(part1) + strlen(dn) + strlen(part3) + strlen(fn1) + strlen(str_sp) + 3;
    if(sz < sizeof(buffer2)) {
      strcpy(buffer2, part1);
      strcat(buffer2, str_dq);
      strcat(buffer2, dn);
      strcat(buffer2, part3);
      strcat(buffer2, str_dq);
      strcat(buffer2, str_sp);
      strcat(buffer2, fn1);
      tracecc_actions[0].ar = buffer2;
    }
  }
}


static
void
modify_path_to_remove_dklibs(void) {
  int have_directory	= 0;	/**< Flag: Have directory name to remove. */
  char dirname[4096];		/**< Base directory to remove from PATH. */
  HKEY reskey;			/**< Registry key. */
  LONG result;			/**< Registry operation result. */
  DWORD	dwType;			/**< Registry entry type. */
  DWORD dwSz;			/**< Registry entry value buffer size. */
  result = RegCreateKeyExA(
    HKEY_LOCAL_MACHINE, "Software\\DkApp", (DWORD)0, NULL,
    REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &reskey, NULL
  );
  if(result == ERROR_SUCCESS) {
    dwType = REG_SZ; dwSz = sizeof(dirname);
    result = RegQueryValueExA(
      reskey, "DIRECTORY", NULL, &dwType, (BYTE *)dirname, &dwSz
    );
    if(result == ERROR_SUCCESS) {
      if(dwType == REG_SZ) {
        if(dwSz > 0) {
	  dirname[(dwSz < sizeof(dirname)) ? dwSz : (sizeof(dirname) - 1)]='\0';
	  have_directory = 1;
	  dksf_correct_fnsep(dirname);
	}
      }
    }
    RegCloseKey(reskey);
  }
  if(have_directory) {
    unregister_directory_in_path_if_necessary(dirname);
  }
}



#if !DKWFTREG_DLL

int main(int argc, char *argv[]) {
  char dirbuffer[1024];
  int cmd = 0;
  int i;
  int value;
  char *directoryname = NULL;
  
#line 1001 "dkwftreg.ctr"

  

  if(argc > 1) {
    if(strcmp(argv[1], "-u") == 0) {
      cmd = 1;
    } else {
      directoryname = argv[1];
      if(argc > 2) {
        install_cmd = 0;
	for(i = 2; i < argc; i++) {
	  if(sscanf(argv[i], "%ld", &value) == 1) {
	    install_cmd |= value;
	  }
	}
      }
    }
  } else {
    install_cmd = 0;
    cmd = install_or_uninstall();
    if(cmd == 0) {
      printf("Dirk Krause's Tools base directory [%s]: ", default_base_dir);
      if(fgets(dirbuffer, sizeof(dirbuffer), stdin)) {
        directoryname = dkstr_start(dirbuffer, NULL);
	if(!(directoryname)) {
	  directoryname = default_base_dir;
	} else {
	  dkstr_chomp(directoryname, NULL);
	}
      } else {
        directoryname = default_base_dir;
      }
      install_cmd |= mod_install_cmd(0, CMD_BMEPS_FILE, 1);
      install_cmd |= mod_install_cmd(1, CMD_FIG2VECT_FILE, 1);
      install_cmd |= mod_install_cmd(2, CMD_TRACECC_DIR, 1);
      install_cmd |= mod_install_cmd(3, CMD_BMEPS_DIR, 1);
      install_cmd |= mod_install_cmd(4, CMD_FIG2VECT_DIR, 0);
    }
  } 
  setup_buffers(directoryname);
  switch(cmd) {
    case -1: {} break;
    case 1: {
      RemoveContextMenuEntries(directoryname);
      modify_path_to_remove_dklibs();
    } break;
    default: {
      register_actions(directoryname);
      register_directory_in_path_if_necessary(directoryname);
    } break;
  }
  
  
#line 1053 "dkwftreg.ctr"

  exit(exval); return exval;
}

#endif


#ifndef LINT
static char sccs_id[] = {
"@(#)dkwftreg.ctr 1.11 02/03/09"
};
#endif

/* vim: set ai sw=2 : */

