

#include "dk.h"
#include "dkmem.h"
#include "dksfc.h"
#include "dksf.h"
#include "dkstr.h"



#include <stdio.h>

#if DK_HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if DK_HAVE_PROCESS_H
#include <process.h>
#endif
#if DK_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if DK_HAVE_PWD_H
#include <pwd.h>
#endif
#if DK_HAVE_GRP_H
#include <grp.h>
#endif




#line 34 "jhelp.ctr"




static char *command_names[] = {
  /*  0 */	"platform",
  NULL
};



static char *platform_strings[] = {
#if DK_HAVE_FNCASEINS
  "filenames.case-sensitive no",
#else
  "filenames.case-sensitive yes",
#endif
#if DK_HAVE_DOS_DRIVE_LETTER
  "filenames.driveletter yes",
#else
  "filenames.driveletter no",
#endif
#if DK_HAVE_FEATURE_BACKSLASH
  "filenames.backslash yes",
#else
  "filenames.backslash no",
#endif
#if DK_HAVE_FEATURE_ALLDOTALL
  "filenames.alldotall yes",
#else
  "filenames.alldotall no",
#endif
#if DK_HAVE_FEATURE_FN83
  "filenames.dos-style yes",
#else
  "filenames.dos-style no",
#endif
#if DK_HAVE_DOTFILENAMES
  "filenames.dotfiles yes",
#else
  "filenames.dotfiles no",
#endif
#if DK_WORDS_BIGENDIAN
  "bigendian yes",
#else
  "bigendian no",
#endif
  NULL
};



static char *keys_for_user[] = {
  "directory.home",		/*  0 */
  "directory.temp",		/*  1 */
  "user.name",			/*  2 */
  "user.uid",			/*  3 */
  "user.gid",			/*  4 */
  "user.gecos",			/*  5 */
  "HOME",			/*  6 */
  "LOGNAME",			/*  7 */
  "TEMP",			/*  8 */
  "TMPDIR",			/*  9 */
  "/tmp",			/* 10 */
  "directory.current",		/* 11 */
  NULL
};



static
void
print_home_directory DK_P0() {


#if VERSION_BEFORE_2008_06_30


  unsigned char found = 0x00;
#if DK_HAVE_GETUID
  {
    uid_t myuid;
    struct passwd *pw;
    myuid = getuid();
#if DK_HAVE_GETPWUID
    pw = getpwuid(myuid);
    if(pw) {
      found = 0x01;
      if(pw->pw_dir) { printf("%s %s\n", keys_for_user[0], pw->pw_dir); }
      if(pw->pw_name) { printf("%s %s\n", keys_for_user[2], pw->pw_name); }
      printf("%s %lu\n", keys_for_user[3], (unsigned long)(pw->pw_uid));
      printf("%s %lu\n", keys_for_user[4], (unsigned long)(pw->pw_gid));
      if(pw->pw_gecos) { printf("%s %s\n", keys_for_user[5], pw->pw_gecos); }
    }
#else
#if DK_HAVE_GETPWENT
#if DK_HAVE_SETPWENT
    setpwent();
#endif
    while((pw = getpwent()) && (found == 0x00)) {
      if(pw->pw_uid == myuid) {
        found = 0x01;
	if(pw->pw_dir) { printf("%s %s\n", keys_for_user[0], pw->pw_dir); }
        if(pw->pw_name) { printf("%s %s\n", keys_for_user[2], pw->pw_name); }
	printf("%s %lu\n", keys_for_user[3], (unsigned long)(pw->pw_uid));
	printf("%s %lu\n", keys_for_user[4], (unsigned long)(pw->pw_gid));
        if(pw->pw_gecos) { printf("%s %s\n", keys_for_user[5], pw->pw_gecos); }
      }
    }
#if DK_HAVE_ENDPWENT
    endpwent();
#endif
#endif
#endif
  }
#else
  /* ##### ! DK_HAVE_GETUID */
#endif
  if(!found) {
    char *ph;
    ph = getenv(keys_for_user[6]);
    if(ph) {
      printf("%s %s\n", keys_for_user[0], ph);
    }
    ph = getenv(keys_for_user[7]);
    if(ph) {
      printf("%s %s\n", keys_for_user[2], ph);
    }
  }


#else

  size_t mpl;
  char *mybuffer;
  long l;
  mpl = (size_t)dksf_get_maxpathlen();
  if(mpl < 512) mpl = 512; if(mpl > 4096) mpl = 4096;
  mybuffer = dk_new(char,mpl);
  if(mybuffer) {
    /*
      0 home
      2 user name
      3 UID
      4 GID
      5 gecos
    */
    if(dksf_get_uname(mybuffer, mpl)) {
      printf("%s %s\n", keys_for_user[2], mybuffer);
    }
    if(dksf_get_home(mybuffer, mpl)) {
      printf("%s %s\n", keys_for_user[0], mybuffer);
    }
    if(dksf_have_getuid()) {
      l = dksf_getuid();
      printf("%s %ld\n", keys_for_user[3], l);
    }
    if(dksf_have_getgid()) {
      l = dksf_getgid();
      printf("%s %ld\n", keys_for_user[4], l);
    }
    dk_delete(mybuffer);
  }


#endif


}




static
void
print_temp_directory DK_P0() {
  size_t mpl;
  char *mybuffer;
#if VERSION_BEFORE_2008_06_30
  char *ph;
  int i, ft;
  unsigned char found = 0x00;
  dk_stat_t	*stp;
  char *p1, *p2;
#endif

  mpl = dksf_get_maxpathlen();
  if(mpl < 512) mpl = 512;
  if(mpl > 4096) mpl = 4096;
  mybuffer = dk_new(char,mpl);
  if(mybuffer) {

#if VERSION_BEFORE_2008_06_30

    for(i = 0; ((i < 3) && (found == 0x00)); i++) {
      ph = NULL;
      switch(i) {
        case 0: ph = getenv(keys_for_user[8]); break;
	case 1: ph = getenv(keys_for_user[9]); break;
	case 2: ph = keys_for_user[10]; break;
      }
      if(ph) {
        if(strlen(ph) < mpl) {
	  strcpy(mybuffer, ph);
	  p1 = mybuffer; p2 = NULL;
	  while((p1) && (found == 0x00)) {
	    p2 = dkstr_chr(p1, ';');
	    if(p2) { *(p2++) = '\0'; }
	    stp = dkstat_open(p1);
	    if(stp) {
	      ft = dkstat_filetype(stp);
	      dkstat_close(stp);
	      if((ft & (~(DK_FT_SYMLINK))) == DK_FT_DIR) {
	        found = 0x01;
		printf("%s %s\n", keys_for_user[1], p1);
	      }
	    }
	    p1 = p2;
	  }
	}
      }
    }
#if DK_HAVE_GETCWD
    if(getcwd(mybuffer, mpl-1)) {
      printf("%s %s\n", keys_for_user[11], mybuffer);
    }
#else
#if DK_HAVE_GETWD
    if(getwd(mybuffer)) {
      printf("%s %s\n", keys_for_user[11], mybuffer);
    }
#endif
#endif

#else

    if(dksf_get_tempdir(mybuffer, mpl)) {
      printf("%s %s\n", keys_for_user[1], mybuffer);
    }
    if(dksf_getcwd(mybuffer, mpl)) {
      printf("%s %s\n", keys_for_user[11], mybuffer);
    }

#endif

    dk_delete(mybuffer);
  }
}




static
void
platform_information DK_P0() {
  char **ptr;
  ptr = platform_strings;
  while(*ptr) { printf("%s\n", *(ptr++)); }
  print_temp_directory();
  print_home_directory();
}



#if DK_HAVE_PROTOTYPES
int main(int argc, char *argv[])
#else
int main(argc, argv); int argc; char *argv[];
#endif
{
  int cmd = 0;
  if(argc > 0) {
    cmd = dkstr_array_index(command_names, argv[1], 1);
    if(cmd < 0) cmd = 0;
  }
  switch(cmd) {
    case 0: { platform_information(); } break;
  }
  exit(0); return 0;
}


#ifndef LINT
static  char sccs_id[] = { "@(#)jhelp.ctr 1.6 02/03/09" };
#endif

