#include "dkbif.h"
#include "dkbifi.h"
#include "dk.h"
#include "dkstr.h"

#line 5 "jpgreport.ctr"



static char buffer[4096];


static int sof_codes[] = {
  DKBIF_I_SOF0,
  DKBIF_I_SOF1,
  DKBIF_I_SOF2,
  DKBIF_I_SOF3,
  0,
  DKBIF_I_SOF5,
  DKBIF_I_SOF6,
  DKBIF_I_SOF7,
  DKBIF_I_SOF8,
  DKBIF_I_SOF9,
  DKBIF_I_SOF10,
  DKBIF_I_SOF11,
  0,
  DKBIF_I_SOF13,
  DKBIF_I_SOF14,
  DKBIF_I_SOF15,
};
static size_t n_sof_codes = sizeof(sof_codes)/sizeof(int);

static
void
run_for(char *filename)
{
  FILE *f;
  dk_bif_jfif_analyze_t ja;
  size_t i;
  int have_a_sof;
  
  f = fopen(filename, "rb");
  if(f) {
    /* ja.is_printing = 0; */
    ja.inputfile = f;
    ja.os = NULL;
    ja.a = NULL;
    dkbif_jfif_analyze(&ja);
    
    printf("%s\n", filename);
    switch(ja.is_jfif) {
      case 0: {
        printf("\tJFIF file shows no problems.\n");
      } break;
      case 1: {
        printf("\tJFIF file contains problematic SOF chunk(s).\n");
      } break;
      case 2: {
      	printf("\tNot a JFIF file\n");
      } break;
    }
    switch(ja.is_jfif) {
      case 0: case 1: case 2: {
        printf("\tJFIF version:       %u.%u\n",
	  (unsigned)(ja.vers_major), (unsigned)(ja.vers_minor)
	);
	printf("\tImages in file:     %lu\n", ja.n_images);
	printf("\tSOF entries:        %lu\n", ja.n_sofs);
	printf("\tSOF(s):             ");
	have_a_sof = 0;
	for(i = 0; i < n_sof_codes; i++) {
	  if(dkbif_jfif_get_sof(ja.used_sofs, sof_codes[i])) {
	    if(have_a_sof) {
	      printf(", ");
	    }
	    have_a_sof = 1;
	    printf("%u", (unsigned)i);
	  }
	}
	printf("\n");
	printf("\tWidth:              %u\n", ja.width);
	printf("\tHeight:             %u\n", ja.height);
	printf("\tComponents:         %u\n", (unsigned)(ja.components));
	printf("\tBits per component: %u\n", (unsigned)(ja.bpc));
	printf("\tDpi setting         %u\n", (unsigned)(ja.dpi_setting));
	printf("\tX resolution:       %u\n", ja.xres);
	printf("\tY resolution:       %u\n", ja.yres);
      } break;
    }
    fclose(f); f = NULL;
  } else {
    printf("%s: ERROR: Failed to open file\n", filename);
  }
  
}


int main(int argc, char *argv[])
{
  int i;
  char *p;
  
#line 100 "jpgreport.ctr"

  if(argc > 1) {
    for(i = 1; i < argc; i++) {
      run_for(argv[i]);
    }
  } else {
    while(fgets(buffer, sizeof(buffer), stdin)) {
      p = dkstr_start(buffer, NULL);
      if(p) {
        dkstr_chomp(p, NULL);
	run_for(p);
      }
    }
  }
  
#line 114 "jpgreport.ctr"

}


#ifndef LINT
static char sccs_id[] = {
"@(#)jpgreport.ctr 1.2 02/03/09\tbmeps2"
};
#endif

