package DKrause::Application;

use 5.001001;
use strict;
use warnings;
use Carp;

require Exporter;
use AutoLoader;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use DKrause::Application ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw(
	
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
	
);

our $VERSION = '0.01';

sub AUTOLOAD {
    # This AUTOLOAD is used to 'autoload' constants from the constant()
    # XS function.

    my $constname;
    our $AUTOLOAD;
    ($constname = $AUTOLOAD) =~ s/.*:://;
    croak "&DKrause::Application::constant not defined" if $constname eq 'constant';
    my ($error, $val) = constant($constname);
    if ($error) { croak $error; }
    {
	no strict 'refs';
	# Fixed between 5.005_53 and 5.005_61
#XXX	if ($] >= 5.00561) {
#XXX	    *$AUTOLOAD = sub () { $val };
#XXX	}
#XXX	else {
	    *$AUTOLOAD = sub { $val };
#XXX	}
    }
    goto &$AUTOLOAD;
}

require XSLoader;
XSLoader::load('DKrause::Application', $VERSION);

# Preloaded methods go here.

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
=head1 NAME

DKrause::Application - Perl extension for using D. Krause's dk_app_t
data type.

=head1 SYNOPSIS

  use DKrause::Application;
  my $a = DKrause::Application::new($an, $gn, $scd, $silently, $filter);
  $a->log($p, $msg);
  $x = $a->getPreference($k, $ex);
  $x = $a->getPreferenceBoolean($k);
  $x = $a->getPreferenceUnsignedLong($k);
  $x = $a->getPreferenceString($k);
  $a->setPreference($k, $v);
  $x = $a->transformString($s);
  $x = $a->transformStringOrFilename($s, $fn);
  $x = $a->findConfigurationFile($n);
  $x = $a->findDataFile($n);
  $x = $a->findString($t, $k, $d);
  $s = $a->readDataFile($fn);
  $s = $a->readConfigurationFile($fn);
  $s = $a->writeFile($fn);
  $s = $a->openFile($fn, $m, $t);
  $x = $a->tempFileName();
  $x = $a->getApplicationName();
  $a->setSilent($f);
  $a->setFilter($f);
  $a->setGroupName($n);
  $a->setSourceFileName($fn);
  $a->setSourceLineNumber($l);
  $x = $a->getSourceFileName();
  $x = $a->getSourceLineNumber();
  $a->getMinimalLogLevel();
  $a->unconfigure();
  $a->setKeepTemporaryFiles($f);
  $l = $a->getMaxPathLength();
  $l = $a->getMaxFiles();
  $x = $a->getcwd();
  $x = $a->getExecutableFile($p);
  $x = $a->combinePaths($s,$t);
  $a->isAbsolutePath($f);
  $x = $a->getFileNameSuffix($s);
  $x = $a->getLastFileName($s);
  $a->correctFileName($s);
  $a->removeDirectory($f);
  $a->removeFile($f);

=head1 DESCRIPTION

This modules provides an object-oriented  Perl interface to the dk_app_t data
type representing an application.

The new() method creates a new dk_app_t structure and returns a reference to
it. The I<an> argument is the application name (string), I<gn> is the
application group name (string). I<scd> is the system configuration directory
to use (string), typically ``/etc'' or ``/usr/local/etc''.
The I<silently> argument is a boolean flag to indicate whether or not
to force the application to run silently (do not print any messages to
standard output or standard error output). The I<filter> argument
is a boolean flag indicating whether the application runs as filter. A filter
prints the normal output to standard output, so we should print error
messages... to standard error output only.

The getApplicationName() returns the name of the application.

The log() method writes a log messages. The I<p> argument is an integer
describing the messages priority: 0 for ``do not print message'', 1
for panic messages, 2 for fatal errors (no recovery possible),
3 for errors (recovery probably possible), 4 for warnings, 5 for
informational messages, 6 for progresses and 7 for debug messages.

The getPreference() method retrieves an application preference,
the arguments are a key I<k> (string) and an integer value telling
the application which preference sources to exclude: 1 for
command line options, 2 for settings made in the program itself,
4 for user-specific settings, 8 for system-wide settings.
Or-combinations can be used to exclude multiple sources. The default
value 0 allows the application to use all available preference
sources.

The getPreferenceBoolean(), getPreferenceString() and
getPreferenceUnsignedLong() methods retrieve a preference, all
available preferences sources are used. The result is converted
into the specified data type.

The setPreference() method assigns a value I<v> (string) to the
specified key I<k> (string).

The transformString() method transforms a string, several
placeholders are replaced by values:

=over	4

=item	$(app.name)

The name of the current application.

=item	$(user.name)

The user name.

=item	$(user.home)

The users home directory.

=item	$(user.uid)

The users UID.

=item	$(user.gid)

The users GID.

=item	$(user.euid)

The users effective UID.

=item	$(user.egid)

The users effective GID.

=item	$(host.name)

The host name (short host name).

=item	$(host.domain)

The hosts DNS domain.

=item	$(app.dir)

The application specific directory (where to find resources, i.e.
string tables, images...).

=item	$(shared.dir)

The shared directory of the application (contains resources, i.e.
string tables, images... used by multiple applications).

=item	$(temp.dir)

The temporary directory specific for this process.

=item	$(process.pid)

The process ID (PID) of the current process.

=item	$(process.ppid)

The PID of the parent process.

=item	$(process.pgid)

The process group ID.

=back

Use a backslash in front of a dollar sign to prevent the method
from substitutions.

The transformStringOrFilename() method does the same transformations for
strings and file names. The I<fn> boolean flag indicates, whether or not
I<s> is a file name.

The findConfigurationFile() method searches for a configuration file
and returns the full file name (path name).

The findDataFile() method searches for a data file and returns the full
file name.

The findString() method searches for a string, I<t> is the name of a
string table, I<k> is the key to search for and I<d> is a default string
to return if the function fails to find a value.

The readDataFile() method searches for a data file, opens the file
and returns a DKrause::Stream object which can be used to read the
file. The object should be closed using $s->close() after the
file contents was read.

The readConfigurationFile() method searches for a configuration file,
opens the file and returns a stream object which can be used to
read the file. Use $s->close() after the contents was read.

The writeFile() method opens a file for writing and returns a stream
object which can be used to write to the file. Use
$s->close() to close the file after all contents was written.

The openFile() method opens a file, I<m> contains the open mode
in text form (the same as for the fopen() runtime function).
The optional I<t> argument is either "plain", "gzip" or "bzip2"
and indicates whether to read/write the contents
``as is'' (plain, the default) or gzip- or bzip2-compressed.

The tempFileName() method returns the full path name of a
temporary file. This file resides in a process-specific temporary
directory which is normally removed when the application object
is destroyed.

The setSilent() method sets the application to silent behaviour
(no messages to standard output or standard error output).
The I<f> boolean flag turns this behaviour on or off.

The setFilter() method sets the application to filter mode
(no messages to standard output).
The I<f> boolean flag turns this behaviour on or off.

The setGroupName() method sets the name of the application group.

The setSourceFileName() method sets the name of the source file
to be printed in error messages.

The setSourceLineNumber() method sets the line number to be printed
in error messages.

The getSourceFileName() and getSourceLineNumber() methods can be used to
retrieve the file name and line number for error messages.

The getMinimalLogLevel() method returns the minimal log level required
to produce any message (to standard output, standard error output, log file
or syslog service). Intended use: If the program does not print any
debug or progress message, we do not need to load the according string
tables.

The unconfigure() method tells the application to remove all 
application-specific preferences for the current user when the application
object is destroyed.

The setKeepTemporaryFiles() method tells the application not to remove
the process-specific temporary directory when the application object
is destroyed. In specific situations one might want to inspect the
contents of the temporary files after the application was finished
for debugging purposes.

The getMaxPathLength() method returns the maximum number of characters
in a path name.

The getMaxFiles() method returns the maximum number of open files.

The getcwd() method returns the current working directory.

The getExecutableFile() method returns the full path name of the
binary executed if the command ``$p'' is run.

The combinePaths() method appends the path name $t to the path $s
and returns the result. If $t is an absolute path, the result
is $t.

The isAbsolutePath() method checks whether or not $f is an absolute
path name.

The getFileNameSuffix() method returns the file type suffix of the path name $s
including the dot (i.e. ``.txt'').

The getLastFileName() method returns the last path component of a path name.

The correctFileName() method corrects all path component separators
(the method switches slashes to backslashes or vice versa).

The removeDirectory() method removes the specified directory.

The removeFile() method removes the specified file.

=head1 EXPORT

None by default.

=head1 AUTHOR

Dirk Krause

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2007 by Dirk Krause

All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

=over	4

=item	Redistributions of source code must retain the above
copyright notice, this list of conditions and the
following disclaimer.

=item	Redistributions in binary form must reproduce the above 
opyright notice, this list of conditions and the following
disclaimer in the documentation and/or other materials
provided with the distribution.

=item	Neither the name of the Dirk Krause nor the names of
contributors may be used to endorse or promote
products derived from this software without specific
prior written permission.

=back

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

=cut

