package DKrause::BitMatrix;

use 5.001001;
use strict;
use warnings;
use Carp;

require Exporter;
use AutoLoader;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use DKrause::BitMatrix ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw(
	
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
	
);

our $VERSION = '0.01';

sub AUTOLOAD {
    # This AUTOLOAD is used to 'autoload' constants from the constant()
    # XS function.

    my $constname;
    our $AUTOLOAD;
    ($constname = $AUTOLOAD) =~ s/.*:://;
    croak "&DKrause::BitMatrix::constant not defined" if $constname eq 'constant';
    my ($error, $val) = constant($constname);
    if ($error) { croak $error; }
    {
	no strict 'refs';
	# Fixed between 5.005_53 and 5.005_61
#XXX	if ($] >= 5.00561) {
#XXX	    *$AUTOLOAD = sub () { $val };
#XXX	}
#XXX	else {
	    *$AUTOLOAD = sub { $val };
#XXX	}
    }
    goto &$AUTOLOAD;
}

require XSLoader;
XSLoader::load('DKrause::BitMatrix', $VERSION);

# Preloaded methods go here.

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
=head1 NAME

DKrause::BitMatrix - Perl extension for bit matrices

=head1 SYNOPSIS

  use DKrause::BitMatrix;
  $m = DKrause::BitMatrix::new($columns,$rows);
  $m->set($x,$y);
  $m->reset($x,$y);
  $res = $m->get($x,$y);
  $res = $m->expand();

=head1 DESCRIPTION

This module allows to store a matrix of bits. The matrix is represented
by a Perl object.

The new() function creates a new object representing a matrix of bits,
the number of columns and rows are specified as arguments.

The set() and reset() methods can be used to set a bit at the specified
position to 1 or 0.

The get() method retrieves a bit from the specified position.

The expand() method can be used to find indirect paths in a graph.
If you have a directed graph (I<n> nodes), create a matrix of I<n> columns
and I<n> rows. For each known direct path from node I<i> to node I<j>
set the bit at position I<x=i>, I<y=j>.
After running expand() retrieve the bit at position I<x=k>, I<y=l> to see
whether or not there is any path (direct or indirect) from node I<k> to node
I<l>.

=head2 EXPORT

None by default.

=head1 AUTHOR

Dirk Krause

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2007 by Dirk Krause

All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

=over	4

=item	*
Redistributions of source code must retain the above
copyright notice, this list of conditions and the
following disclaimer.

=item	*
Redistributions in binary form must reproduce the above 
opyright notice, this list of conditions and the following
disclaimer in the documentation and/or other materials
provided with the distribution.

=item	*
Neither the name of the Dirk Krause nor the names of
contributors may be used to endorse or promote
products derived from this software without specific
prior written permission.

=back

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

=cut
