package DKrause::BitStream;

use 5.001001;
use strict;
use warnings;
use Carp;

require Exporter;
use AutoLoader;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use DKrause::BitStream ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw(
	
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
	
);

our $VERSION = '0.01';

sub AUTOLOAD {
    # This AUTOLOAD is used to 'autoload' constants from the constant()
    # XS function.

    my $constname;
    our $AUTOLOAD;
    ($constname = $AUTOLOAD) =~ s/.*:://;
    croak "&DKrause::BitStream::constant not defined" if $constname eq 'constant';
    my ($error, $val) = constant($constname);
    if ($error) { croak $error; }
    {
	no strict 'refs';
	# Fixed between 5.005_53 and 5.005_61
#XXX	if ($] >= 5.00561) {
#XXX	    *$AUTOLOAD = sub () { $val };
#XXX	}
#XXX	else {
	    *$AUTOLOAD = sub { $val };
#XXX	}
    }
    goto &$AUTOLOAD;
}

require XSLoader;
XSLoader::load('DKrause::BitStream', $VERSION);

# Preloaded methods go here.

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__

=head1 NAME

DKrause::BitStream - Perl extension to write bits to streams.

=head1 SYNOPSIS

  use DKrause::Stream;
  use DKrause::BitStream;

  $s = DKrause::BitStream::new($stream);
  $s->put($value, $nbits);
  $s->flush();
  $s->close();

=head1 DESCRIPTION

This module provides a class to write single bits into stream 
or output/encoding filter objects.

=over	4

=item	* DKrause::BitStream::new($stream)

creates a new bit-stream object and returns a reference to it.

The objects resources should be released using the close()
method as soon as the object is no longer needed.

=item	* close()

releases resources assigned to the object.

If close() is not called explicitly the resources are released
during object desctruction.

=item	* put($value, $nbits)

shifts the B<last> n bits of value into the stream.

=item	* flush()

flushes output. If there are bits in the internal buffer but
less than 8, a byte is written to the output stream, the last
bits are 0-padded.

=back

=head1 EXPORT

None by default.

=head1 EXAMPLE

Imagine you have a three-color bitmap image file using 4 bits per component
and you want to export this to EPS level 3.

  use DKrause::Stream;
  use DKrause::BitStream;

  my $s = DKrause::Stream::new('output.eps', 'w');
  if($s) {
    # ... write start of EPS file ...
    # now write image data
    my $f = $s->filter("ascii85", "flate");
    if($f) {
      if($f->startChunk()) {
        my $bf = DKrause::BitStream::new($f);
        if($bf) {
          my $w = getWidth();	# get image width
	  my $h = getHeight();	# get height
	  my $x; my $y;
	  my $r; my $g; my $b;
	  for($y = 0; $y < $h; $y++) {
	    for($x = 0; $x < $w; $x++) {
	      $r = getRed($x, $y);	# get red value for pixel
	      $bf->put($r, 4);
	      $g = getGreen($x, $y);	# get green value for pixel
	      $bf->put($g, 4);
	      $b = getBlue($x, $y);	# get blue value for pixel
	      $bf->put($b, 4);
	    }
	    $bf->flush();
	  }
          $bf->close();
        }
        $f->endChunk();
      }
      $f->close();
    }
    # image data written
    # ... write remaining text of EPS file ...
    $s->close();
  }


=head1 SEE ALSO

DKrause::Stream

=head1 AUTHOR

Dirk Krause

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2007 by Dirk Krause

All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

=over	4

=item	Redistributions of source code must retain the above
copyright notice, this list of conditions and the
following disclaimer.

=item	Redistributions in binary form must reproduce the above 
opyright notice, this list of conditions and the following
disclaimer in the documentation and/or other materials
provided with the distribution.

=item	Neither the name of the Dirk Krause nor the names of
contributors may be used to endorse or promote
products derived from this software without specific
prior written permission.

=back

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

=cut
