package DKrause::Stream;

use 5.001001;
use strict;
use warnings;
use Carp;

require Exporter;
use AutoLoader;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use DKrause::Stream ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw(
	
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
	
);

our $VERSION = '0.01';

sub AUTOLOAD {
    # This AUTOLOAD is used to 'autoload' constants from the constant()
    # XS function.

    my $constname;
    our $AUTOLOAD;
    ($constname = $AUTOLOAD) =~ s/.*:://;
    croak "&DKrause::Stream::constant not defined" if $constname eq 'constant';
    my ($error, $val) = constant($constname);
    if ($error) { croak $error; }
    {
	no strict 'refs';
	# Fixed between 5.005_53 and 5.005_61
#XXX	if ($] >= 5.00561) {
#XXX	    *$AUTOLOAD = sub () { $val };
#XXX	}
#XXX	else {
	    *$AUTOLOAD = sub { $val };
#XXX	}
    }
    goto &$AUTOLOAD;
}

require XSLoader;
XSLoader::load('DKrause::Stream', $VERSION);

# Preloaded methods go here.

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
=head1 NAME

DKrause::Stream - Perl extension for output streams and
output/encoding filters

=head1 SYNOPSIS

  use DKrause::Stream;

  $s1 = DKrause::Stream::new($filename, $mode);
  $s2 = DKrause::Stream::newStdout();
  $s3 = DKrause::Stream::newStderr();
  $s1->nl();
  $s1->crnl();
  $s1->puts($string);
  $s1->putsDouble($value);
  $s1->putsLong($value);
  $s1->putsUnsignedLong($value);
  $value = $s1->gets();
  $s1->wbByte($value);
  $s1->wbWord($value);
  $s1->wbUword($value);
  $s1->wbDword($value);
  $s1->wbUdword($value);
  $s1->wbString($string);
  $value = $s1->rbByte();
  $value = $s1->rbWord();
  $value = $s1->rbUword();
  $value = $s1->rbDword();
  $value = $s1->rbUdword();
  $string = $s1->rbString();
  $value = $s1->getBytesWritten();
  $s4 = $s1->filter( list_of_methods );
  $s4->setCrNl($boolvalue);
  $s4->setFinalizing($boolvalue);
  $boolvalue = $s4->startChunk();
  $boolvalue = $s4->endChunk();
  $s1->close(); # s1 and s2 too


=head1 DESCRIPTION

=head2 OVERVIEW

The DKrause::Stream module provides an object class for
input/output streams and output encoding/compression filters.

Some file formats consist of both plain text and chunks
of binary compressed and encoded data. This module can be
used to produce such files.

=head2 STREAM API

The stream API consists of functions to create and close
streams, to read and write data and to create filter objects
on top of the stream.

The following constructors can be used to create stream objects:

=over	4

=item	* DKrause::Stream::new($filename, $mode)

opens a file in the specified mode (same as for the C fopen() function),
creates a Perl stream object to access the file and returns the
object reference.

After the file is written or read completely one should use the returned
objets close() method as soon as possible to close the file.

If close() is not used, the file is closed during object destruction.

=item	* DKrause::Stream::newStdout()

creates a stream object to access standard output.

=item	* DKrause::Stream::newStderr()

creates a stream object to access standard error output.

=back

The following methods are available for stream objects:

=over	4

=item	* nl()

writes a newline to the stream. This should be used for files
opened in text mode. The translation of newlines to
carriage return/newline combinations -- if necessary -- is left
up to the operating system.

=item	* crnl()

write a carriage return/newline combination to the stream. This
should be used for files opened in binary mode.

=item	* puts($text)

writes text to the stream. No newline is added.

=item	* putsDouble($value)

converts a numeric (double) value to text and writes that text
to the stream.

=item	* putsLong($value)

converts an integer value to text and writes the text to the
stream.

=item	* putsUnsignedLong($value)

converts a long unsigned value to text and write the text to
the stream.

=item	* $value = $stream->gets()

reads a text line from the stream and returns the line
(including the newline).
The function returns undef if no line can be read.

=item	* wbByte($value)

writes one byte to the stream.

=item	* wbWord($value)

writes one two-byte word to the stream. The word is written binary
in network representation (MSB first, LSB last).

=item	* wbUword($value)

write one unsigned two-byte word in network representation to
the stream.

=item	* wbDword($value)

write one double-word (four bytes) to the stream in network
representation.

=item	* wbUdword($value)

writes on unsigned double-word to the stream in network representation.

=item	* wbString($text)

write the string to the stream in 2 parts:

=over	8

=item	the string length (including the finishing 0 byte) as
two-byte word in network representation and

=item	the string itself, including the finishing 0 byte.

=back

=item	* $value = $stream->rbByte()

reads a byte from the stream and returns the value or undef.

=item	* $value = $stream->rbWord()

reads a word from the stream and returns the value (converted
back to host representation) or undef.

=item	* $value = $stream->rbUword()

... unsigned word...

=item	* $value = $stream->rbDword()

... double-word ...

=item	* $value = $stream->rbUdword()

... unsigned double-word ...

=item	* $value = $stream->rbString()

reads a string and returns the value or undef.

=item	* $value = $stream->getBytesWritten()

returns the number of bytes written to the stream.

=item	* $filterobject = $stream->filter( list_of_methods );

creates a filter object on top of the stream and returns the
new object by reference. The list_of_methods contains
keywords which compression/encoding mechanisms to use, you
can choose from the following list: "run-length", "ascii85",
"asciihex", "flate" and "buffered".
The lowest level filter must be mentioned first. If -- i.e.
you want to flate compress and ASCII85-encode -- you have
to write

  $filterobject = $stream->filter("ascii85", "flate");

=back

=head2 ADDITIONAL FILTER API

The API for filter objects is the same as for stream objects,
extensions are added to start and end a compressed/encoded
chunk, and to control some compression/encoding details.

The following methods are available at filter objects in addition
to the methods from the stream API:

=over	4

=item	* setCrNl($boolvalue)

switches between the use of carriage return/newline combination
($boolvalue is true) or the use of simple newlines.

=item	* setFinalizing($boolvalue)

turns encoding chunk finalization of ($boolvalue = true) or off.

If chunk finalization is on, a ``>'' is appended to
ASCII-Hex-encoded output chunks, ``~>'' is appended to ASCII-85-encoded
output chunks.

=item	* startChunk()

starts a compressed/encoded chunk.

=item	* endChunk()

finishes a compressed/encoded chunk and flushes all internal buffers.

=back

=head1 EXPORT

Non exports by default.

=head1 EXAMPLE

The following example will create a PDF file showing the german
flag. It shows how to use streams and compression/encoding output
filters on top of a stream.

The stream $spdf for the PDF file is created using
DKrause::Stream::new(). On top of this stream we have
a compression/encoding filter $sc, created by invoking the
filter() method of $spdf.
The lowest encoding layer (used immediately before writing
to file) is ASCII85, this is mentioned first. On top of
it there is a compression layer for flate compression.

Each time we want to add compressed/encoded output we have to
call the startChunk() and endChunk() methods of $sc before
we start and after we have finished compressed output.

For PDF files we have to create an xref table in the end, listing
the file positions where objects start. So at the beginning of each
object we use the getBytesWritten() method to obtain the current
file position, this is stored in the @objstart array.

  use DKrause::Stream;
  
  my @objstart;
  my $nextobj = 1;
  $objstart[0] = 0;
  
  sub start_object($)
  {
    my $s = shift;
    $objstart[$nextobj] = $s->getBytesWritten();
    $s->puts("$nextobj 0 obj");
    $s->crnl();
    $nextobj++;
  }
  
  sub end_object($)
  {
    my $s = shift;
    $s->puts("endobj"); $s->crnl();
  }
  
  sub finish_pdf_file($)
  {
    my $s = shift;
    my $i = 0;
    my $pos = $s->getBytesWritten();
    $s->puts("xref"); $s->crnl();
    $s->puts("0 $nextobj"); $s->crnl();
    $s->puts("0000000000 65535 f");
    $s->crnl();
    for($i = 1; $i < $nextobj; $i++) {
      $s->puts(sprintf("\%010d 00000 n", $objstart[$i]));
      $s->crnl();
    }
    $s->puts("trailer"); $s->crnl();
    $s->puts("<</Size $nextobj /Info 1 0 R /Root 2 0 R>>");
    $s->crnl();
    $s->puts("startxref"); $s->crnl();
    $s->puts("$pos"); $s->crnl();
    $s->puts("\%\%EOF"); $s->crnl();
  }
  
  # create stream for PDF file
  my $spdf = DKrause::Stream::new('flag.pdf', 'wb');
  if($spdf) {
    # create encoding filter on top of spdf
    my $sc = $spdf->filter('ascii85', 'flate');
    if($sc) {
      $sc->setCrNl(1);
      $spdf->puts("\%PDF-1.4"); $spdf->crnl();
      # info object (1)
      start_object($spdf);
      $spdf->puts("<</Producer (the PDF creation example)>>");
      $spdf->crnl();
      end_object($spdf);
      # root object (2)
      start_object($spdf);
      $spdf->puts("<</Type /Catalog /Pages 3 0 R>>");
      $spdf->crnl();
      end_object($spdf);
      # pages tree object (3)
      start_object($spdf);
      $spdf->puts("<</Type /Pages /Kids [4 0 R] /Count 1>>");
      $spdf->crnl();
      end_object($spdf);
      # page object (4)
      start_object($spdf);
      $spdf->puts("<<"); $spdf->crnl();
      $spdf->puts("/Type /Page"); $spdf->crnl();
      $spdf->puts("/Parent 3 0 R"); $spdf->crnl();
      $spdf->puts("/MediaBox [0 0 144 108]"); $spdf->crnl();
      $spdf->puts("/CropBox [0 0 144 108]"); $spdf->crnl();
      $spdf->puts("/Contents 5 0 R"); $spdf->crnl();
      $spdf->puts("/Rotate 0"); $spdf->crnl();
      $spdf->puts("/Resources <</ProcSet [/PDF]>>"); $spdf->crnl();
      $spdf->puts(">>"); $spdf->crnl();
      end_object($spdf);
      # content stream object (5)
      start_object($spdf);
      $spdf->puts("<</Length 6 0 R /Filter [/ASCII85Decode /FlateDecode]>>");
      $spdf->crnl();
      $spdf->puts("stream"); $spdf->crnl();
      my $startpos = $spdf->getBytesWritten();
      if($sc->startChunk()) {
        # save graphics state
        $sc->puts("q"); $sc->crnl();
        # background
        $sc->puts("0 0 m 144 0 l 144 108 l 0 108 l h W n");
        $sc->crnl();
        # rectangle 1: yellow rectangle at bottom
        $sc->puts("1 0.839 0 rg"); $sc->crnl();
        $sc->puts("0 0 m 144 0 l 144 36 l 0 36 l h f*");
        $sc->crnl();
        # rectangle 2: red rectangle in the middle
        $sc->puts("0.82 0 0 rg"); $sc->crnl();
        $sc->puts("0 36 m 144 36 l 144 72 l 0 72 l h f*");
        $sc->crnl();
        # rectangle 3: black rectangle at top
        $sc->puts("0 0 0 rg"); $sc->crnl();
        $sc->puts("0 72 m 144 72 l 144 108 l 0 108 l h f*");
        $sc->crnl();
        # restore graphics state
        $sc->puts("Q"); $sc->crnl();
        $sc->endChunk();
      }
      $spdf->crnl();
      my $endpos = $spdf->getBytesWritten();
      $spdf->puts("endstream"); $spdf->crnl();
      end_object($spdf);
      # content length object (6)
      start_object($spdf);
      $endpos = $endpos - $startpos;
      $spdf->puts("$endpos"); $spdf->crnl();
      end_object($spdf);
      # issue trailer and end of file...
      finish_pdf_file($spdf);
      # close encoding filter
      $sc->close();
    }
    # close output file
    $spdf->close();
  }

=head1 AUTHOR

Dirk Krause

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2007 by Dirk Krause

All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

=over	4

=item	Redistributions of source code must retain the above
copyright notice, this list of conditions and the
following disclaimer.

=item	Redistributions in binary form must reproduce the above 
opyright notice, this list of conditions and the following
disclaimer in the documentation and/or other materials
provided with the distribution.

=item	Neither the name of the Dirk Krause nor the names of
contributors may be used to endorse or promote
products derived from this software without specific
prior written permission.

=back

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

=cut
