/*
Copyright (c) 2000-2005, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#include "dk.h"
#include "dkstr.h"
#include "dkmem.h"
#include "dksf.h"

#include <stdio.h>
#include <string.h>
#include <winbase.h>
#include <winreg.h>

struct _key_val_ {
  char *key;
  char *value;
} ;
typedef struct _key_val_ key_val;

static key_val keys_to_set[] = {
  { "all/all:/ui/lang",    NULL },
  { "all/all:/dir/app",    NULL },
  { "all/all:/dir/shared", NULL },
  { "all/all:/dir/tmp",    NULL },
  { "all/all:/log/file/level", "warning" },
  { "all/all:/log/file/keep",  "warning" },
  { "all/all:/log/file/name",  NULL },
  { "all/all:/log/stderr/level", "error" },
  { "all/all:/log/stdout/level", "none" },
  { "all/all:/log/stdout/codepage", NULL },
  { "all/all:/log/stderr/codepage", NULL },
  { "all/all:/snmp/log/file/name", NULL },
  { NULL, NULL }
};

static char shared_suffix[] = { "/lib" };
static char app_suffix[] = { "/lib/$(app.name)" };
static char cp850[] = { "cp.850" };
static char deflf[] = { "/$(app.name).$(process.pid).log" };
static char snmplf[] = { "/snmp.$(process.pid).log" };

static char *appdir = NULL;
static char *shareddir = NULL;
static char *tmpdir = NULL;
static char *buffer = NULL;
static char *logfile = NULL;
static char *snmpfile = NULL;


#line 83 "winsetup.ctr"


static long maxpathlen;

static void correct_string(char *str)
{
  char *ptr;
  ptr = str;
  while(*ptr) {
    if(*ptr == '\\') *ptr = '/';
    ptr++;
  }
}

static void save_the_keys(void)
{
  HKEY hkTemp;
  PHKEY phkSubkey;
  DWORD disp, dwType, sz;
  LONG retval;
  key_val *kvptr;
  int ok;
  hkTemp = 0;
  phkSubkey = &hkTemp;
  retval = RegCreateKeyExA(
    HKEY_LOCAL_MACHINE, "Software\\Dkapp",
    0, NULL, REG_OPTION_NON_VOLATILE,
    KEY_ALL_ACCESS,
    NULL,
    phkSubkey,
    &disp
  );
  if(retval == ERROR_SUCCESS) { 
    kvptr = keys_to_set;
    while(kvptr->key) { 
      if(kvptr->value) { 
	ok = 0; retval = ERROR_SUCCESS - 1L;
	sz = maxpathlen;
	phkSubkey = &hkTemp;
        retval = RegQueryValueExA(
	  hkTemp,
	  kvptr->key,
	  NULL,
	  &dwType,
	  buffer,
	  &sz
	);
	if(retval == ERROR_SUCCESS) { 
	  if(dwType == REG_SZ) { 
	    ok = 1;
	  }
	}
	if(!ok) { 
	  retval = RegSetValueExA(
	    hkTemp,
	    kvptr->key,
	    0,
	    REG_SZ,
	    kvptr->value,
	    (strlen(kvptr->value) + 1)
	  );
	  
	}
      }
      kvptr++;
    }
    RegCloseKey(hkTemp);
  }
}

int main(int argc, char *argv[])
{
  
#line 155 "winsetup.ctr"

  if(argc == 3) { 
    keys_to_set[0].value = argv[2];
    maxpathlen = dksf_get_maxpathlen();
    appdir = dk_new(char,maxpathlen);
    shareddir = dk_new(char,maxpathlen);
    tmpdir = dk_new(char,maxpathlen);
    buffer = dk_new(char,maxpathlen);
    logfile = dk_new(char,maxpathlen);
    snmpfile = dk_new(char,maxpathlen);
    if(appdir && shareddir && tmpdir && buffer && logfile && snmpfile) {
      if((strlen(argv[1]) + strlen(app_suffix)) < (size_t)maxpathlen) {
        strcpy(appdir, argv[1]);
        strcat(appdir, app_suffix);
        keys_to_set[1].value = appdir;
	
      }
      if((strlen(argv[1]) + strlen(shared_suffix)) < (size_t)maxpathlen) {
        strcpy(shareddir, argv[1]);
        strcat(shareddir, shared_suffix);
        keys_to_set[2].value = shareddir;
	
      }
      if(dksf_get_tempdir(tmpdir,maxpathlen)) {
	keys_to_set[3].value = tmpdir;
	if((strlen(tmpdir) + strlen(deflf)) < (size_t)maxpathlen) {
	  strcpy(logfile, tmpdir);
	  strcat(logfile, deflf);
	  keys_to_set[6].value = logfile;
	  correct_string(logfile);
	}
	if((strlen(tmpdir) + strlen(snmplf)) < (size_t)maxpathlen) {
	  strcpy(snmpfile, tmpdir);
	  strcat(logfile, snmplf);
	  keys_to_set[11].value = snmpfile;
	  correct_string(snmpfile);
	}
	
      }
      if(strcmp(argv[2], "de") == 0) {
	keys_to_set[9].value = cp850;
	keys_to_set[10].value = cp850;
      }
      if(keys_to_set[1].value) correct_string(keys_to_set[1].value);
      if(keys_to_set[2].value) correct_string(keys_to_set[2].value);
      if(keys_to_set[3].value) correct_string(keys_to_set[3].value);
      save_the_keys();
    }
    if(snmpfile) { dk_delete(snmpfile); }
    snmpfile = NULL;
    if(appdir) { dk_delete(appdir); }
    appdir = NULL;
    if(shareddir) { dk_delete(shareddir); }
    shareddir = NULL;
    if(tmpdir) { dk_delete(tmpdir); }
    tmpdir = NULL;
    if(buffer) { dk_delete(buffer); }
    buffer = NULL;
    if(logfile) { dk_delete(logfile); }
    logfile = NULL;
  }
  
#line 216 "winsetup.ctr"

  exit(0); return 0;
}



#ifndef LINT
static char sccs_id[] = {
"@(#)winsetup.ctr 1.13 02/03/09\t(krause) - Initial registry setup for windows"
};
#endif

